from collections.abc import Iterable, Mapping, Set
from datetime import date, datetime
from typing import Any, ClassVar, cast

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.resource_classes.annotation import AnnotationResponse, AssetLinkData, FileLinkData
from cognite_toolkit._cdf_tk.client.resource_classes.asset import AssetResponse
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    DirectNodeRelation,
    EdgeReference,
    EdgeRequest,
    InstanceSource,
    NodeReference,
    NodeRequest,
    ViewCorePropertyResponse,
    ViewReference,
    ViewResponseProperty,
)
from cognite_toolkit._cdf_tk.client.resource_classes.event import EventResponse
from cognite_toolkit._cdf_tk.client.resource_classes.filemetadata import FileMetadataResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, InternalId
from cognite_toolkit._cdf_tk.client.resource_classes.legacy.migration import AssetCentricId
from cognite_toolkit._cdf_tk.client.resource_classes.resource_view_mapping import ResourceViewMappingRequest
from cognite_toolkit._cdf_tk.client.resource_classes.timeseries import TimeSeriesResponse
from cognite_toolkit._cdf_tk.utils.collection import flatten_dict_json_path
from cognite_toolkit._cdf_tk.utils.dtype_conversion import (
    asset_centric_convert_to_primary_property,
    convert_to_primary_property,
)
from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentricTypeExtended
from cognite_toolkit._cdf_tk.utils.useful_types2 import AssetCentricResourceExtended

from .data_model import COGNITE_MIGRATION_SPACE_ID, INSTANCE_SOURCE_VIEW_ID
from .issues import ConversionIssue, FailedConversion, InvalidPropertyDataType


class DirectRelationCache:
    """Cache for direct relation references to look up target of direct relations.

    This is used when creating direct relations from asset-centric resources to assets, files, and source systems.
    """

    class TableName:
        ASSET_ID = "assetId"
        SOURCE_NAME = "source"
        FILE_ID = "fileId"
        ASSET_EXTERNAL_ID = "assetExternalId"
        FILE_EXTERNAL_ID = "fileExternalId"

    ASSET_ID_PROPERTIES: ClassVar[Set[tuple[str, str]]] = {
        ("timeseries", "assetId"),
        ("file", "assetIds"),
        ("event", "assetIds"),
        ("sequence", "assetId"),
        ("annotation", "data.assetRef.id"),
        ("asset", "parentId"),
    }
    SOURCE_NAME_PROPERTIES: ClassVar[Set[tuple[str, str]]] = {
        ("asset", "source"),
        ("event", "source"),
        ("file", "source"),
    }
    FILE_ID_PROPERTIES: ClassVar[Set[tuple[str, str]]] = {
        ("annotation", "data.fileRef.id"),
        ("annotation", "annotatedResourceId"),
    }
    ASSET_EXTERNAL_ID_PROPERTIES: ClassVar[Set[tuple[str, str]]] = {("annotation", "data.assetRef.externalId")}
    FILE_EXTERNAL_ID_PROPERTIES: ClassVar[Set[tuple[str, str]]] = {("annotation", "data.fileRef.externalId")}

    def __init__(self, client: ToolkitClient) -> None:
        self._client = client
        self._cache_map: dict[tuple[str, str] | str, dict[str, NodeReference] | dict[int, NodeReference]] = {}
        # Constructing the cache map to be accessed by both table name and property id
        for table_name, properties in [
            (self.TableName.ASSET_ID, self.ASSET_ID_PROPERTIES),
            (self.TableName.SOURCE_NAME, self.SOURCE_NAME_PROPERTIES),
            (self.TableName.FILE_ID, self.FILE_ID_PROPERTIES),
            (self.TableName.ASSET_EXTERNAL_ID, self.ASSET_EXTERNAL_ID_PROPERTIES),
            (self.TableName.FILE_EXTERNAL_ID, self.FILE_EXTERNAL_ID_PROPERTIES),
        ]:
            cache: dict[str, NodeReference] | dict[int, NodeReference] = {}
            self._cache_map[table_name] = cache
            for key in properties:
                self._cache_map[key] = cache

    def update(self, resources: Iterable[AssetCentricResourceExtended]) -> None:
        """Update the cache with direct relation references for the given asset-centric resources.

        This is used to bulk update the cache for a chunk of resources before converting them to data model instances.
        """
        asset_ids: set[int] = set()
        source_ids: set[str] = set()
        file_ids: set[int] = set()
        asset_external_ids: set[str] = set()
        file_external_ids: set[str] = set()
        for resource in resources:
            if isinstance(resource, AnnotationResponse):
                if resource.annotated_resource_type == "file" and resource.annotated_resource_id:
                    file_ids.add(resource.annotated_resource_id)
                self._extract_annotation_refs(resource.data, asset_ids, asset_external_ids, file_ids, file_external_ids)
            elif isinstance(resource, AssetResponse):
                if resource.source:
                    source_ids.add(resource.source)
                if resource.parent_id is not None:
                    asset_ids.add(resource.parent_id)
            elif isinstance(resource, FileMetadataResponse):
                if resource.source:
                    source_ids.add(resource.source)
                if resource.asset_ids:
                    asset_ids.update(resource.asset_ids)
            elif isinstance(resource, EventResponse):
                if resource.source:
                    source_ids.add(resource.source)
                if resource.asset_ids:
                    asset_ids.update(resource.asset_ids)
            elif isinstance(resource, TimeSeriesResponse):
                if resource.asset_id is not None:
                    asset_ids.add(resource.asset_id)
        if asset_ids:
            self._update_cache(self._client.migration.lookup.assets(id=list(asset_ids)), self.TableName.ASSET_ID)
        if source_ids:
            # SourceSystems are not cached in the client, so we have to handle the caching ourselves.
            cache = cast(dict[str, NodeReference], self._cache_map[self.TableName.SOURCE_NAME])
            missing: dict[str, str] = {}
            for source_id in source_ids:
                if source_id.casefold() not in cache:
                    missing[source_id.casefold()] = source_id
                elif source_id not in cache:
                    missing[source_id] = source_id
            if missing:
                source_systems = self._client.migration.created_source_system.retrieve(list(missing))
                for source_system in source_systems:
                    source_reference = source_system.as_direct_relation_reference()
                    cache[source_system.source] = source_reference
                    if original_str := missing.get(source_system.source):
                        cache[original_str] = source_reference
        if file_ids:
            self._update_cache(self._client.migration.lookup.files(id=list(file_ids)), self.TableName.FILE_ID)
        if asset_external_ids:
            self._update_cache(
                self._client.migration.lookup.assets(external_id=list(asset_external_ids)),
                self.TableName.ASSET_EXTERNAL_ID,
            )
        if file_external_ids:
            self._update_cache(
                self._client.migration.lookup.files(external_id=list(file_external_ids)),
                self.TableName.FILE_EXTERNAL_ID,
            )

    @staticmethod
    def _extract_annotation_refs(
        data: AssetLinkData | FileLinkData | dict[str, Any],
        asset_ids: set[int],
        asset_external_ids: set[str],
        file_ids: set[int],
        file_external_ids: set[str],
    ) -> None:
        if isinstance(data, AssetLinkData):
            if isinstance(data.asset_ref, InternalId):
                asset_ids.add(data.asset_ref.id)
            elif isinstance(data.asset_ref, ExternalId):
                asset_external_ids.add(data.asset_ref.external_id)
        elif isinstance(data, FileLinkData):
            if isinstance(data.file_ref, InternalId):
                file_ids.add(data.file_ref.id)
            elif isinstance(data.file_ref, ExternalId):
                file_external_ids.add(data.file_ref.external_id)

    def _update_cache(
        self, instance_id_by_id: dict[int, NodeReference] | dict[str, NodeReference], table_name: str
    ) -> None:
        cache = self._cache_map[table_name]
        for identifier, instance_id in instance_id_by_id.items():
            cache[identifier] = NodeReference(space=instance_id.space, external_id=instance_id.external_id)  # type: ignore[index]

    def get_cache(
        self, resource_type: AssetCentricTypeExtended, property_id: str
    ) -> Mapping[str | int, NodeReference] | None:
        """Get the cache for the given resource type and property ID."""
        return self._cache_map.get((resource_type, property_id))  # type: ignore[return-value]


def asset_centric_to_dm(
    resource: AssetCentricResourceExtended,
    instance_id: NodeReference | EdgeReference,
    view_source: ResourceViewMappingRequest,
    view_properties: dict[str, ViewResponseProperty],
    direct_relation_cache: DirectRelationCache,
    preferred_consumer_view: ViewReference | None = None,
) -> tuple[NodeRequest | EdgeRequest | None, ConversionIssue]:
    """Convert an asset-centric resource to a data model instance.

    Args:
        resource (CogniteResource): The asset-centric resource to convert.
        instance_id (NodeId | EdgeApply): The ID of the instance to create or update.
        view_source (ResourceViewMappingRequest): The view source defining how to map the resource to the data model.
        view_properties (dict[str, ViewProperty]): The defined properties referenced in the view source mapping.
        direct_relation_cache (DirectRelationCache): Cache for direct relation references.
        preferred_consumer_view (ViewId | None): The preferred consumer view for the instance.

    Returns:
        tuple[NodeApply | EdgeApply, ConversionIssue]: A tuple containing the converted NodeApply and any ConversionIssue encountered.
    """
    resource_type = _lookup_resource_type(resource)
    dumped = resource.dump()
    try:
        id_ = dumped.pop("id")
    except KeyError as e:
        raise ValueError("Resource must have an 'id' field.") from e
    if not isinstance(id_, int):
        raise TypeError(f"Resource 'id' field must be an int, got {type(id_)}.")
    data_set_id = dumped.pop("dataSetId", None)
    external_id = dumped.pop("externalId", None)

    issue = ConversionIssue(
        id=str(AssetCentricId(resource_type, id_=id_)),
        asset_centric_id=AssetCentricId(resource_type, id_=id_),
        instance_id=NodeReference(space=instance_id.space, external_id=instance_id.external_id),
    )

    properties = create_properties(
        dumped,
        view_properties,
        view_source.property_mapping,
        resource_type,
        issue=issue,
        direct_relation_cache=direct_relation_cache,
    )
    sources: list[InstanceSource] = []
    if properties:
        sources.append(
            InstanceSource(
                source=ViewReference(
                    space=view_source.view_id.space,
                    external_id=view_source.view_id.external_id,
                    version=view_source.view_id.version,
                ),
                properties=properties,
            )
        )

    if resource_type != "annotation":
        instance_source_properties = {
            "resourceType": resource_type,
            "id": id_,
            "dataSetId": data_set_id,
            "classicExternalId": external_id,
            "resourceViewMapping": {"space": COGNITE_MIGRATION_SPACE_ID, "externalId": view_source.external_id},
        }
        if preferred_consumer_view:
            instance_source_properties["preferredConsumerViewId"] = {
                **preferred_consumer_view.dump(),
                "type": "view",
            }
        sources.append(
            InstanceSource(
                source=ViewReference(
                    space=INSTANCE_SOURCE_VIEW_ID.space,
                    external_id=INSTANCE_SOURCE_VIEW_ID.external_id,
                    version=view_source.view_id.version,
                ),
                properties=instance_source_properties,
            )
        )

    instance: NodeRequest | EdgeRequest
    if isinstance(instance_id, EdgeReference):
        edge_properties = create_edge_properties(
            dumped, view_source.property_mapping, resource_type, issue, direct_relation_cache, instance_id.space
        )
        if any(key not in edge_properties for key in ("start_node", "end_node", "type")):
            # Failed conversion of edge properties
            return None, issue
        instance = EdgeRequest(
            space=instance_id.space,
            external_id=instance_id.external_id,
            sources=sources,
            **edge_properties,  # type: ignore[arg-type]
        )
    elif isinstance(instance_id, NodeReference):
        instance = NodeRequest(space=instance_id.space, external_id=instance_id.external_id, sources=sources)
    else:
        raise RuntimeError(f"Unexpected instance_id type {type(instance_id)}")

    return instance, issue


def _lookup_resource_type(resource_type: AssetCentricResourceExtended) -> AssetCentricTypeExtended:
    if isinstance(resource_type, AssetResponse):
        return "asset"
    elif isinstance(resource_type, FileMetadataResponse):
        return "file"
    elif isinstance(resource_type, EventResponse):
        return "event"
    elif isinstance(resource_type, TimeSeriesResponse):
        return "timeseries"
    elif isinstance(resource_type, AnnotationResponse):
        if resource_type.annotated_resource_type == "file" and resource_type.annotation_type in (
            "diagrams.AssetLink",
            "diagrams.FileLink",
        ):
            return "annotation"
    raise ValueError(f"Unsupported resource type: {resource_type}")


def create_properties(
    dumped: dict[str, Any],
    view_properties: dict[str, ViewResponseProperty],
    property_mapping: dict[str, str],
    resource_type: AssetCentricTypeExtended,
    issue: ConversionIssue,
    direct_relation_cache: DirectRelationCache,
) -> dict[str, JsonValue]:
    """
    Create properties for a data model instance from an asset-centric resource.

    Args:
        dumped: Dict representation of the asset-centric resource.
        view_properties: Defined properties referenced in the view source mapping.
        property_mapping:  Mapping from asset-centric property IDs to data model property IDs.
        resource_type: The type of the asset-centric resource (e.g., "asset", "timeseries").
        issue: ConversionIssue object to log any issues encountered during conversion.
        direct_relation_cache: Cache for direct relation references to look up target of direct relations.

    Returns:
        Dict of property IDs to PropertyValueWrite objects.

    """
    flatten_dump = flatten_dict_json_path(dumped, keep_structured=set(property_mapping.keys()))
    properties: dict[str, JsonValue] = {}
    ignored_asset_centric_properties: set[str] = set()
    for prop_json_path, prop_id in property_mapping.items():
        if prop_json_path not in flatten_dump:
            continue
        if prop_id not in view_properties:
            continue
        if prop_id in properties:
            ignored_asset_centric_properties.add(prop_json_path)
            continue
        dm_prop = view_properties[prop_id]
        if not isinstance(dm_prop, ViewCorePropertyResponse):
            issue.invalid_instance_property_types.append(
                InvalidPropertyDataType(property_id=prop_id, expected_type=ViewCorePropertyResponse.__name__)
            )
            continue
        try:
            value = asset_centric_convert_to_primary_property(
                flatten_dump[prop_json_path],
                dm_prop.type,
                dm_prop.nullable or False,
                destination_container_property=(dm_prop.container, dm_prop.container_property_identifier),
                source_property=(resource_type, prop_json_path),
                direct_relation_lookup=direct_relation_cache.get_cache(resource_type, prop_json_path),
            )
        except (ValueError, TypeError, NotImplementedError) as e:
            issue.failed_conversions.append(
                FailedConversion(property_id=prop_json_path, value=flatten_dump[prop_json_path], error=str(e))
            )
            continue
        if isinstance(value, date):
            # Convert date to ISO format string, as the data model expects dates as strings in ISO format.
            properties[prop_id] = value.isoformat()
        elif isinstance(value, datetime):
            # Convert datetime to ISO format string, as the data model expects datetimes as strings in ISO format.
            properties[prop_id] = value.isoformat(timespec="milliseconds")
        else:
            properties[prop_id] = value

    issue.ignored_asset_centric_properties = sorted(
        (set(flatten_dump.keys()) - set(property_mapping.keys())) | ignored_asset_centric_properties
    )
    issue.missing_asset_centric_properties = sorted(set(property_mapping.keys()) - set(flatten_dump.keys()))
    # Node and edge properties are handled separately
    issue.missing_instance_properties = sorted(
        {
            prop_id
            for prop_id in property_mapping.values()
            if not (prop_id.startswith("edge.") or prop_id.startswith("node."))
        }
        - set(view_properties.keys())
    )
    return properties


def create_edge_properties(
    dumped: dict[str, Any],
    property_mapping: dict[str, str],
    resource_type: AssetCentricTypeExtended,
    issue: ConversionIssue,
    direct_relation_cache: DirectRelationCache,
    default_instance_space: str,
) -> dict[str, NodeReference]:
    flatten_dump = flatten_dict_json_path(dumped)
    edge_properties: dict[str, NodeReference] = {}
    for prop_json_path, prop_id in property_mapping.items():
        if not prop_id.startswith("edge."):
            continue
        if prop_json_path not in flatten_dump:
            continue
        edge_prop_id = prop_id.removeprefix("edge.")
        if edge_prop_id in ("startNode", "endNode", "type"):
            value: NodeReference | Any
            # DirectRelation lookup.
            try:
                value = convert_to_primary_property(
                    flatten_dump[prop_json_path],
                    DirectNodeRelation(),
                    False,
                    direct_relation_lookup=direct_relation_cache.get_cache(resource_type, prop_json_path),
                )
            except (ValueError, TypeError, NotImplementedError) as e:
                issue.failed_conversions.append(
                    FailedConversion(property_id=prop_json_path, value=flatten_dump[prop_json_path], error=str(e))
                )
                continue
        elif edge_prop_id.endswith(".externalId"):
            # Just an external ID string.
            edge_prop_id = edge_prop_id.removesuffix(".externalId")
            value = NodeReference(space=default_instance_space, external_id=str(flatten_dump[prop_json_path]))
        else:
            issue.invalid_instance_property_types.append(
                InvalidPropertyDataType(property_id=prop_id, expected_type="EdgeProperty")
            )
            continue
        # We know that value is DirectRelationReference here
        edge_properties[edge_prop_id.replace("Node", "_node")] = value  # type: ignore[assignment]

    return edge_properties



from abc import ABC
from typing import Annotated, Literal

from pydantic import Field, TypeAdapter

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject


class IndexDefinition(BaseModelObject, ABC):
    index_type: str
    properties: list[str]


class BtreeIndex(IndexDefinition):
    index_type: Literal["btree"] = "btree"
    by_space: bool | None = None
    cursorable: bool | None = None


class InvertedIndex(IndexDefinition):
    index_type: Literal["inverted"] = "inverted"


Index = Annotated[BtreeIndex | InvertedIndex, Field(discriminator="index_type")]

IndexAdapter: TypeAdapter[Index] = TypeAdapter(Index)



from collections.abc import Iterable, Sequence
from typing import Any, Literal

from pydantic import JsonValue, TypeAdapter

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.hosted_extractor_source import (
    HostedExtractorSourceRequestUnion,
    HostedExtractorSourceResponse,
    HostedExtractorSourceResponseUnion,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId


class HostedExtractorSourcesAPI(CDFResourceAPI[HostedExtractorSourceResponseUnion]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/hostedextractors/sources", item_limit=10),
                "retrieve": Endpoint(method="POST", path="/hostedextractors/sources/byids", item_limit=100),
                "update": Endpoint(method="POST", path="/hostedextractors/sources/update", item_limit=10),
                "delete": Endpoint(method="POST", path="/hostedextractors/sources/delete", item_limit=100),
                "list": Endpoint(method="GET", path="/hostedextractors/sources", item_limit=100),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[HostedExtractorSourceResponseUnion]:
        if isinstance(response, SuccessResponse):
            data = response.body_json
        else:
            data = TypeAdapter(dict[str, JsonValue]).validate_json(response.body)
        items = [HostedExtractorSourceResponse.validate_python(item) for item in data.get("items", [])]
        return PagedResponse[HostedExtractorSourceResponseUnion](items=items, nextCursor=data.get("nextCursor"))

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[ExternalId]:
        return ResponseItems[ExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[HostedExtractorSourceRequestUnion]) -> list[HostedExtractorSourceResponseUnion]:
        """Create hosted extractor sources in CDF.

        Args:
            items: List of source request objects to create.
        Returns:
            List of created source response objects.
        """
        return self._request_item_response(list(items), "create")

    def retrieve(
        self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False
    ) -> list[HostedExtractorSourceResponseUnion]:
        """Retrieve hosted extractor sources from CDF.

        Args:
            items: List of ExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved source response objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def update(
        self, items: Sequence[HostedExtractorSourceRequestUnion], mode: Literal["patch", "replace"] = "replace"
    ) -> list[HostedExtractorSourceResponseUnion]:
        """Update hosted extractor sources in CDF.

        Args:
            items: List of source request objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated source response objects.
        """
        return self._update(list(items), mode=mode)

    def delete(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False, force: bool | None = None) -> None:
        """Delete hosted extractor sources from CDF.

        Args:
            items: List of ExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
            force: Delete any jobs associated with each item.
        """
        extra_body: dict[str, Any] = {"ignoreUnknownIds": ignore_unknown_ids}
        if force is not None:
            extra_body["force"] = force
        self._request_no_response(items, "delete", extra_body=extra_body)

    def paginate(
        self,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[HostedExtractorSourceResponseUnion]:
        """Iterate over hosted extractor sources in CDF.

        Args:
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of source response objects.
        """
        return self._paginate(cursor=cursor, limit=limit)

    def iterate(
        self,
        limit: int = 100,
    ) -> Iterable[list[HostedExtractorSourceResponseUnion]]:
        """Iterate over hosted extractor sources in CDF.

        Args:
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of source response objects.
        """
        return self._iterate(limit=limit)

    def list(self, limit: int | None = 100) -> list[HostedExtractorSourceResponseUnion]:
        """List all hosted extractor sources in CDF.

        Args:
            limit: Maximum number of items to return. None for all items.

        Returns:
            List of source response objects.
        """
        return self._list(limit=limit)



from collections.abc import Sequence
from typing import Annotated, Any, Literal

from pydantic import AliasChoices, Field, model_serializer, model_validator

from cognite_toolkit._cdf_tk.client._resource_base import Identifier

SemanticVersion = Annotated[
    str,
    Field(
        min_length=5,
        max_length=14,
        pattern=r"^(0|[1-9]\d{0,3})\.(0|[1-9]\d{0,3})\.(0|[1-9]\d{0,3})$",
    ),
]


class InternalOrExternalIdDefinition(Identifier):
    type: str


class InternalIdUnwrapped(Identifier):
    id: int

    def __str__(self) -> str:
        return f"id={self.id}"

    @model_serializer
    def serialize(self) -> int:
        return self.id

    @model_validator(mode="before")
    @classmethod
    def deserialize(cls, value: Any) -> Any:
        if isinstance(value, int):
            return {"id": value}
        return value


class InternalId(InternalOrExternalIdDefinition):
    type: Literal["id"] = Field("id", exclude=True)
    id: int

    @classmethod
    def from_ids(cls, ids: Sequence[int]) -> list["InternalId"]:
        return [cls(id=id_) for id_ in ids]

    def __str__(self) -> str:
        return f"id={self.id}"

    def as_unwrapped(self) -> InternalIdUnwrapped:
        return InternalIdUnwrapped(id=self.id)


class ExternalId(InternalOrExternalIdDefinition):
    type: Literal["externalId"] = Field("externalId", exclude=True)
    external_id: str

    @classmethod
    def from_external_ids(cls, external_ids: list[str]) -> list["ExternalId"]:
        return [cls(external_id=ext_id) for ext_id in external_ids]

    def __str__(self) -> str:
        return f"externalId='{self.external_id}'"


InternalOrExternalId = Annotated[InternalId | ExternalId, Field(discriminator="type")]


class NameId(Identifier):
    name: str

    def __str__(self) -> str:
        return f"name='{self.name}'"


class RawDatabaseId(Identifier):
    name: str = Field(alias="name", validation_alias=AliasChoices("dbName", "name"))

    def __str__(self) -> str:
        return f"name='{self.name}'"

    def dump(self, camel_case: bool = True, exclude_extra: bool = False) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        Args:
            camel_case (bool): Will be ignored. Included for compatibility.
            exclude_extra (bool): Will be ignored. Included for compatibility.

        """
        return self.model_dump(mode="json", by_alias=False)


class RawTableId(Identifier):
    name: str
    db_name: str

    def __str__(self) -> str:
        return f"{self.db_name}.{self.name}"


class SequenceRowId(Identifier):
    external_id: str = Field(description="ExternalId of the sequence")
    rows: tuple[int, ...]

    def __str__(self) -> str:
        rows_str = ", ".join(str(row) for row in self.rows)
        return f"externalId='{self.external_id}', rows=[{rows_str}]"


class ExtractionPipelineConfigId(Identifier):
    external_id: str
    revision: int | None = None

    def __str__(self) -> str:
        return f"externalId='{self.external_id}', revision={self.revision}"


class WorkflowVersionId(Identifier):
    workflow_external_id: str
    version: str

    def __str__(self) -> str:
        return f"workflowExternalId='{self.workflow_external_id}', version='{self.version}'"


class ThreeDModelRevisionId(Identifier):
    model_id: int = Field(exclude=True)
    id: int

    def __str__(self) -> str:
        return f"modelId={self.model_id}, id={self.id}"


class DataSetId(Identifier):
    data_set_id: int

    def __str__(self) -> str:
        return f"dataSetId={self.data_set_id}"


class DataProductVersionId(Identifier):
    data_product_external_id: str
    version: SemanticVersion

    def __str__(self) -> str:
        return f"dataProductExternalId='{self.data_product_external_id}', version='{self.version}'"


class TransformationNotificationId(Identifier):
    transformation_external_id: str
    destination: str

    def __str__(self) -> str:
        return f"transformationExternalId='{self.transformation_external_id}', destination='{self.destination}'"


class PrincipalId(Identifier):
    id: str

    def __str__(self) -> str:
        return f"id='{self.id}'"


class PrincipalLoginId(Identifier):
    principal: str
    id: str

    def __str__(self) -> str:
        return f"principal='{self.principal}', id='{self.id}'"



from collections.abc import Hashable, Iterable, Sequence
from typing import Any, final

from cognite.client.data_classes.capabilities import Capability, StreamsAcl

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.streams import (
    StreamRequest,
    StreamResponse,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import StreamYAML

from .datamodel import ContainerCRUD


@final
class StreamCRUD(ResourceCRUD[ExternalId, StreamRequest, StreamResponse]):
    folder_name = "streams"
    resource_cls = StreamResponse
    resource_write_cls = StreamRequest
    kind = "Streams"
    yaml_cls = StreamYAML
    dependencies = frozenset({ContainerCRUD})
    _doc_url = "Streams/operation/createStream"
    support_update = False

    @property
    def display_name(self) -> str:
        return "streams"

    @classmethod
    def get_id(cls, item: StreamRequest | StreamResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[StreamRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [StreamsAcl.Action.Read]
            if read_only
            else [StreamsAcl.Action.Read, StreamsAcl.Action.Create, StreamsAcl.Action.Delete]
        )
        return StreamsAcl(actions, StreamsAcl.Scope.All())

    def create(self, items: Sequence[StreamRequest]) -> list[StreamResponse]:
        return self.client.streams.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[StreamResponse]:
        return self.client.streams.retrieve(list(ids), ignore_unknown_ids=True)

    def delete(self, ids: Sequence[ExternalId]) -> int:
        self.client.streams.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[StreamResponse]:
        if data_set_external_id or space or parent_ids:
            # These filters are not supported for Streams
            return iter([])

        all_streams = self.client.streams.list()
        return iter(all_streams)



import traceback
from collections.abc import Sequence
from graphlib import TopologicalSorter
from pathlib import Path

import questionary
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError
from rich import print
from rich.panel import Panel

from cognite_toolkit._cdf_tk.client._resource_base import T_Identifier, T_RequestResource, T_ResponseResource
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.constants import (
    _RUNNING_IN_BROWSER,
    BUILD_ENVIRONMENT_FILE,
    HINT_LEAD_TEXT,
    HINT_LEAD_TEXT_LEN,
)
from cognite_toolkit._cdf_tk.cruds import (
    CRUDS_BY_FOLDER_NAME,
    DataCRUD,
    DataSetsCRUD,
    RawDatabaseCRUD,
    ResourceContainerCRUD,
    ResourceCRUD,
    ResourceWorker,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import Loader
from cognite_toolkit._cdf_tk.cruds._resource_cruds import SimulatorModelRevisionCRUD, SimulatorRoutineRevisionCRUD
from cognite_toolkit._cdf_tk.data_classes import (
    BuildEnvironment,
    DeployResults,
    ResourceContainerDeployResult,
    ResourceDeployResult,
)
from cognite_toolkit._cdf_tk.data_classes._module_directories import ReadModule
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitCleanResourceError,
    ToolkitMissingModuleError,
    ToolkitNotADirectoryError,
    ToolkitValidationError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.tk_warnings import (
    LowSeverityWarning,
    MediumSeverityWarning,
    ToolkitDependenciesIncludedWarning,
    ToolkitNotSupportedWarning,
)
from cognite_toolkit._cdf_tk.utils import (
    humanize_collection,
    read_yaml_file,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables

from ._utils import _print_ids_or_length

AVAILABLE_DATA_TYPES: tuple[str, ...] = tuple(CRUDS_BY_FOLDER_NAME)


class CleanCommand(ToolkitCommand):
    def clean_resources(
        self,
        loader: ResourceCRUD[T_Identifier, T_RequestResource, T_ResponseResource],
        env_vars: EnvironmentVariables,
        read_modules: list[ReadModule],
        dry_run: bool = False,
        drop: bool = True,
        drop_data: bool = False,
        verbose: bool = False,
    ) -> ResourceDeployResult | None:
        if not isinstance(loader, ResourceContainerCRUD) and not drop:
            # Skipping silently as this, we will not drop data or delete this resource
            return ResourceDeployResult(name=loader.display_name)
        if not loader.support_drop:
            print(f"  [bold green]INFO:[/] {loader.display_name!r} cleaning is not supported, skipping...")
            return ResourceDeployResult(name=loader.display_name)
        elif isinstance(loader, ResourceContainerCRUD) and not drop_data:
            print(
                f"  [bold]INFO:[/] Skipping cleaning of {loader.display_name!r}. This is a data resource (it contains "
                f"data and is not only configuration/metadata) and therefore "
                "requires the --drop-data flag to be set to perform cleaning..."
            )
            return ResourceContainerDeployResult(name=loader.display_name, item_name=loader.item_name)

        worker = ResourceWorker(loader, "clean")
        files = worker.load_files(read_modules=read_modules)
        if not files:
            return None
        # Since we do a clean, we do not want to verify that everything exists wrt data sets, spaces etc.
        local_resources = worker.load_resources(
            filepaths=files,
            environment_variables=env_vars.dump(include_os=True),
            is_dry_run=True,
        )
        existing_resources = loader.retrieve(list(local_resources.keys()))
        nr_of_existing = len(existing_resources)

        if drop:
            prefix = "Would clean" if dry_run else "Cleaning"
            with_data = "with data " if isinstance(loader, ResourceContainerCRUD) else ""
        else:
            prefix = "Would drop data from" if dry_run else "Dropping data from"
            with_data = ""
        print(f"[bold]{prefix} {nr_of_existing} {loader.display_name} {with_data}from CDF...[/]")
        if not isinstance(loader, RawDatabaseCRUD):
            for duplicate in worker.duplicates:
                self.warn(LowSeverityWarning(f"Duplicate {loader.display_name} {duplicate}."))

        # Deleting resources.
        if isinstance(loader, ResourceContainerCRUD) and drop_data:
            nr_of_dropped_datapoints = self._drop_data(existing_resources, loader, dry_run, verbose)
            if drop:
                nr_of_deleted = self._delete_resources(existing_resources, loader, dry_run, verbose)
            else:
                nr_of_deleted = 0
            if verbose:
                print("")
            return ResourceContainerDeployResult(
                name=loader.display_name,
                deleted=nr_of_deleted,
                total=nr_of_existing,
                dropped_datapoints=nr_of_dropped_datapoints,
                item_name=loader.item_name,
            )
        elif not isinstance(self, ResourceContainerCRUD) and drop:
            nr_of_deleted = self._delete_resources(existing_resources, loader, dry_run, verbose)
            if verbose:
                print("")
            return ResourceDeployResult(name=loader.display_name, deleted=nr_of_deleted, total=nr_of_existing)
        else:
            return ResourceDeployResult(name=loader.display_name)

    def _delete_resources(
        self, loaded_resources: Sequence[T_ResponseResource], loader: ResourceCRUD, dry_run: bool, verbose: bool
    ) -> int:
        nr_of_deleted = 0
        resource_ids = loader.get_ids(loaded_resources)
        if dry_run:
            nr_of_deleted += len(resource_ids)
            if verbose:
                print(f"  Would have deleted {_print_ids_or_length(resource_ids)}.")
            return nr_of_deleted

        try:
            nr_of_deleted += loader.delete(resource_ids)
        except CogniteAPIError as e:
            self.warn(MediumSeverityWarning(f"Failed to delete {_print_ids_or_length(resource_ids)}. Error {e}."))
            if verbose:
                print(Panel(traceback.format_exc()))
        except CogniteNotFoundError:
            if verbose:
                print(f"  [bold]INFO:[/] {_print_ids_or_length(resource_ids)} do(es) not exist.")
        else:  # Delete succeeded
            if verbose:
                print(f"  Deleted {_print_ids_or_length(resource_ids)}.")
        return nr_of_deleted

    def _drop_data(
        self,
        loaded_resources: Sequence[T_ResponseResource],
        loader: ResourceContainerCRUD,
        dry_run: bool,
        verbose: bool,
    ) -> int:
        nr_of_dropped = 0
        resource_ids = loader.get_ids(loaded_resources)
        if dry_run:
            resource_drop_count = loader.count(resource_ids)
            nr_of_dropped += resource_drop_count
            if verbose:
                self._verbose_print_drop(resource_drop_count, resource_ids, loader, dry_run)
            return nr_of_dropped

        try:
            resource_drop_count = loader.drop_data(resource_ids)
            nr_of_dropped += resource_drop_count
        except CogniteAPIError as e:
            if e.code == 404 and verbose:
                print(f"  [bold]INFO:[/] {len(resource_ids)} {loader.display_name} do(es) not exist.")
        except CogniteNotFoundError:
            return nr_of_dropped
        else:  # Delete succeeded
            if verbose:
                self._verbose_print_drop(resource_drop_count, resource_ids, loader, dry_run)
        return nr_of_dropped

    def _interactive_module_selection(self, built_modules: list[ReadModule] | None) -> list[ReadModule] | None:
        if not built_modules:
            return None
        choices = [
            questionary.Choice(title=built_module.dir.name, value=built_module) for built_module in built_modules
        ]

        selected_modules = questionary.checkbox(
            "Which modules would you like to clean?",
            instruction="Use arrow up/down, press space to select item(s) and enter to save",
            choices=choices,
            validate=lambda choice: "You must select at least one module." if len(choice) == 0 else True,
        ).unsafe_ask()

        if not selected_modules:
            return None
        return selected_modules

    def _verbose_print_drop(
        self, drop_count: int, resource_ids: Sequence[T_Identifier], loader: ResourceContainerCRUD, dry_run: bool
    ) -> None:
        prefix = "Would have dropped" if dry_run else "Dropped"
        if drop_count > 0:
            print(
                f"  {prefix} {drop_count:,} {loader.item_name} from {loader.display_name}: "
                f"{_print_ids_or_length(resource_ids)}."
            )
        elif drop_count == 0:
            verb = "is" if len(resource_ids) == 1 else "are"
            print(
                f"  The {loader.display_name}: {_print_ids_or_length(resource_ids)} {verb} empty, "
                f"thus no {loader.item_name} will be {'touched' if dry_run else 'dropped'}."
            )
        else:
            # Count is not supported
            print(f" {prefix} all {loader.item_name} from {loader.display_name}: {_print_ids_or_length(resource_ids)}.")

    def _select_modules(self, clean_state: BuildEnvironment, module_str: str | None) -> list[ReadModule] | None:
        if module_str:
            return [module for module in clean_state.read_modules if module.dir.name == module_str]
        return self._interactive_module_selection(clean_state.read_modules)

    def execute(
        self,
        env_vars: EnvironmentVariables,
        build_dir: Path,
        build_env_name: str | None,
        dry_run: bool,
        include: list[str] | None,
        module_str: str | None,
        verbose: bool,
        all_modules: bool = False,
    ) -> None:
        if not build_dir.exists():
            raise ToolkitNotADirectoryError(
                "The build directory does not exists. Did you forget to run `cdf-tk build` first?"
            )
        clean_state = BuildEnvironment.load(read_yaml_file(build_dir / BUILD_ENVIRONMENT_FILE), build_env_name, "clean")
        clean_state.set_environment_variables()
        errors = clean_state.check_source_files_changed()
        for error in errors:
            self.warn(error)
        if errors:
            raise ToolkitCleanResourceError(
                "One or more source files have been modified since the last build. Please rebuild the project."
            )
        client = env_vars.get_client(clean_state.is_strict_validation)
        environment_vars = ""
        if not _RUNNING_IN_BROWSER:
            environment_vars = f"\n\nConnected to {env_vars.as_string()}"

        action = ""
        if dry_run:
            action = "(dry-run) "

        print(
            Panel(
                f"[bold]Cleaning {action}[/]resource from CDF project {client.config.project} based "
                f"on resource files in {build_dir} directory."
                f"{environment_vars}",
                expand=False,
            )
        )

        if not build_dir.is_dir():
            raise ToolkitNotADirectoryError(f"'{build_dir}'. Did you forget to run `cdf build` first?")

        selected_modules: list[ReadModule]
        if all_modules:
            selected_modules = clean_state.read_modules or []
            if not selected_modules:
                raise ToolkitValueError("No modules available to clean.")
        elif module_str:
            selected_modules = [module for module in clean_state.read_modules if module.dir.name == module_str]
            if not selected_modules:
                available_module_names = {module.dir.name for module in clean_state.read_modules}
                raise ToolkitMissingModuleError(
                    f"No modules matched the selection: {module_str}. Available modules: {sorted(available_module_names)}"
                )
        else:
            selected_modules = self._interactive_module_selection(clean_state.read_modules) or []
            if not selected_modules:
                raise ToolkitValueError(
                    "No module specified with the --module option and no modules selected interactively."
                )

        selected_resource_folders = {
            resource_folder for module in selected_modules for resource_folder in module.resource_directories
        }
        selected_loaders = self.get_selected_loaders(build_dir, selected_resource_folders, include)

        results = DeployResults([], "clean", dry_run=dry_run)

        resolved_list: list[type[Loader]] = []
        should_include: list[type[Loader]] = []
        # The topological sort can include loaders that are not selected, so we need to check for that.
        for loader_cls in TopologicalSorter(selected_loaders).static_order():
            if loader_cls in selected_loaders:
                resolved_list.append(loader_cls)
            elif (build_dir / loader_cls.folder_name).is_dir():
                should_include.append(loader_cls)
            # Otherwise, it is not in the build directory and not selected, so we skip it.
            # There should be a warning in the build step if it is missing.
        if should_include:
            self.warn(ToolkitDependenciesIncludedWarning([item.folder_name for item in should_include]))

        for loader_cls in reversed(resolved_list):
            if not issubclass(loader_cls, ResourceCRUD):
                continue
            loader = loader_cls.create_loader(client, build_dir)
            if isinstance(loader, DataSetsCRUD | SimulatorModelRevisionCRUD | SimulatorRoutineRevisionCRUD):
                self.warn(ToolkitNotSupportedWarning(feature=f"{loader.display_name} clean."))
                continue
            result = self.clean_resources(
                loader,
                env_vars=env_vars,
                read_modules=selected_modules,
                drop=True,
                dry_run=dry_run,
                drop_data=True,
                verbose=verbose,
            )
            if result:
                results[result.name] = result
        if results.has_counts:
            print(results.counts_table())
        if results.has_uploads:
            print(results.uploads_table())

    def get_selected_loaders(
        self, build_dir: Path, read_resource_folders: set[str], include: list[str] | None
    ) -> dict[type[Loader], frozenset[type[Loader]]]:
        selected_loaders: dict[type[Loader], frozenset[type[Loader]]] = {}
        for folder_name, loader_classes in CRUDS_BY_FOLDER_NAME.items():
            if include is not None and folder_name not in include:
                continue
            if folder_name in read_resource_folders:
                selected_loaders.update({loader_cls: loader_cls.dependencies for loader_cls in loader_classes})
                continue
            if not (build_dir / folder_name).is_dir():
                continue
            folder_has_supported_files = False
            for loader_cls in loader_classes:
                if loader_cls.any_supported_files(build_dir / folder_name):
                    folder_has_supported_files = True
                    selected_loaders[loader_cls] = loader_cls.dependencies
                elif issubclass(loader_cls, DataCRUD):
                    # Data Loaders are always included, as they will have
                    # the files in the module folder and not the build folder.
                    selected_loaders[loader_cls] = loader_cls.dependencies

            if not folder_has_supported_files:
                kinds = [loader_cls.kind for loader_cls in loader_classes]
                yaml_file = next((build_dir / folder_name).glob("*.yaml"), None)
                suggestion = ""
                if yaml_file:
                    suggestion = f"\n{' ' * HINT_LEAD_TEXT_LEN}For example: '{yaml_file.stem}.{kinds[0]}.yaml'."
                self.warn(
                    MediumSeverityWarning(
                        f"No supported files found in {folder_name!r} folder. Skipping...\n"
                        f"{HINT_LEAD_TEXT}All resource in the {folder_name!r} folder are expected to have suffix: "
                        f"{humanize_collection(kinds)!r}.{suggestion}"
                    )
                )
        return selected_loaders

    @staticmethod
    def validate_include(include: list[str] | None) -> list[str]:
        if include and (invalid_types := set(include).difference(AVAILABLE_DATA_TYPES)):
            raise ToolkitValidationError(
                f"Invalid resource types specified: {invalid_types}, available types: {AVAILABLE_DATA_TYPES}"
            )
        include = include or list(AVAILABLE_DATA_TYPES)
        return include



from functools import cached_property
from pathlib import Path

from pydantic import BaseModel, ConfigDict, Field, JsonValue

from cognite_toolkit._cdf_tk.constants import MODULES

from ._insights import InsightList
from ._module import ModuleSource
from ._types import AbsoluteDirPath, RelativeDirPath, RelativeFilePath, ValidationType


class BuildParameters(BaseModel):
    organization_dir: Path
    build_dir: Path = Field(default_factory=lambda: Path.cwd() / "build")
    config_yaml_name: str | None = Field(
        None,
        description="The name of the configuration YAML file to use. It expected to be"
        "named config.[name].yaml and be located in the organization directory.",
    )
    user_selected_modules: list[str] | None = Field(
        None,
        description="List of module names or paths to build. If not provided, Toolkit will first attempt to find a config YAML "
        "and the modules specified there. If no config YAML is found, Toolkit will build all modules in the organization directory.",
    )

    @property
    def modules_directory(self) -> Path:
        return self.organization_dir / MODULES


class BuildSourceFiles(BaseModel):
    """Intermediate format used when parsing modules"""

    yaml_files: list[RelativeFilePath] = Field(
        description="List of all YAML files that are part of the build, with paths relative to the organization directory."
    )
    selected_modules: set[RelativeDirPath | str] = Field(
        description="Set of modules to build. Either module names (folder names) or relative paths to the organization directory."
    )
    variables: dict[str, JsonValue] = Field(
        default_factory=dict, description="Variables to be used during the build process."
    )
    validation_type: ValidationType = "prod"
    cdf_project: str
    organization_dir: AbsoluteDirPath


class BuiltModule(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    source: ModuleSource
    built_files: list[Path] = Field(default_factory=list)
    insights: InsightList = Field(default_factory=InsightList)

    @property
    def resource_by_type(self) -> dict[str, dict[str, list[Path]]]:
        """Organizes built files by their resource type."""
        resource_by_type: dict[str, dict[str, list[Path]]] = {}
        for file in self.built_files:
            resource_type = file.stem.split(".")[-1]
            resource_type_folder = file.parent.name
            resource_by_type.setdefault(resource_type_folder, {}).setdefault(resource_type, []).append(file)

        return resource_by_type

    @property
    def is_success(self) -> bool:
        return True if self.built_files else False


class BuildLineage(BaseModel): ...


class BuildFolder(BaseModel):
    """Built folder acts as a container holding all built modules and insights from the build process."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    path: Path
    built_modules: list[BuiltModule] = Field(default_factory=list)

    @cached_property
    def insights(self) -> InsightList:
        """Insights from all built modules."""
        insights = InsightList()
        for module in self.built_modules:
            insights.extend(module.insights)
        return insights

    @property
    def lineage(self) -> BuildLineage:
        """Lineage should be generated based on the built modules, but for now it is just a placeholder."""
        return BuildLineage()

    @property
    def built_modules_by_success(self) -> dict[bool, list[str]]:
        """Organizes built modules by their success status."""
        modules_by_success: dict[bool, list[str]] = {True: [], False: []}
        for built_module in self.built_modules:
            modules_by_success[built_module.is_success].append(built_module.source.name)

        return modules_by_success



from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import SimulatorModelRoutineRevisionFilter
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalOrExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.simulator_routine_revision import (
    SimulatorRoutineRevisionRequest,
    SimulatorRoutineRevisionResponse,
)


class SimulatorRoutineRevisionsAPI(CDFResourceAPI[SimulatorRoutineRevisionResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/simulators/routines/revisions", item_limit=1),
                "retrieve": Endpoint(method="POST", path="/simulators/routines/revisions/byids", item_limit=1),
                "list": Endpoint(method="POST", path="/simulators/routines/revisions/list", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[SimulatorRoutineRevisionResponse]:
        return PagedResponse[SimulatorRoutineRevisionResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalOrExternalId]:
        return ResponseItems[InternalOrExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[SimulatorRoutineRevisionRequest]) -> list[SimulatorRoutineRevisionResponse]:
        """Create simulator routine revisions in CDF.

        Args:
            items: List of SimulatorRoutineRevisionRequest objects to create.

        Returns:
            List of created SimulatorRoutineRevisionResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False
    ) -> list[SimulatorRoutineRevisionResponse]:
        """Retrieve simulator routine revisions from CDF.

        Args:
            items: List of InternalOrExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.

        Returns:
            List of retrieved SimulatorRoutineRevisionResponse objects.
        """
        if ignore_unknown_ids:
            return self._request_item_split_retries(items, method="retrieve")
        else:
            return self._request_item_response(items, method="retrieve")

    def paginate(
        self,
        filter: SimulatorModelRoutineRevisionFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[SimulatorRoutineRevisionResponse]:
        """Iterate over simulator routine revisions in CDF.

        Args:
            filter: Filter to apply to the list of simulator routine revisions.
            limit: Maximum number of items to return per page.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of SimulatorRoutineRevisionResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            body={
                "filter": filter.dump() if filter else None,
            },
        )

    def iterate(
        self,
        filter: SimulatorModelRoutineRevisionFilter | None = None,
        limit: int = 100,
    ) -> Iterable[list[SimulatorRoutineRevisionResponse]]:
        """Iterate over simulator routine revisions in CDF.

        Args:
            filter: Filter to apply to the list of simulator routine revisions.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of SimulatorRoutineRevisionResponse objects.
        """
        return self._iterate(
            limit=limit,
            body={
                "filter": filter.dump() if filter else None,
            },
        )

    def list(
        self,
        limit: int | None = 100,
    ) -> list[SimulatorRoutineRevisionResponse]:
        """List all simulator routine revisions in CDF.

        Args:
            limit: Maximum number of items to return. None for all items.

        Returns:
            List of SimulatorRoutineRevisionResponse objects.
        """
        return self._list(limit=limit)



from pathlib import Path
from typing import Annotated, Literal, TypeAlias

from pydantic import PlainValidator


def _is_relative_file_path(p: Path) -> Path:
    if not isinstance(p, Path):
        # Let pydantic handle the type error.
        return p
    if not p.suffix:
        raise ValueError(f"{p.as_posix()!r} is not a file.")
    if p.is_absolute():
        raise ValueError(f"{p.as_posix()!r} is not a relative path.")
    return p


def _is_absolute_file_path(p: Path) -> Path:
    if not isinstance(p, Path):
        # Let pydantic handle the type error.
        return p
    if not p.suffix:
        raise ValueError(f"{p.as_posix()!r} is not a file.")
    if not p.is_absolute():
        raise ValueError(f"{p.as_posix()!r} is not an absolute path.")
    return p


def _is_relative_dir_path(p: Path) -> Path:
    if not isinstance(p, Path):
        # Let pydantic handle the type error.
        return p
    if p.suffix:
        raise ValueError(f"{p.as_posix()!r} is not a directory.")
    if p.is_absolute():
        raise ValueError(f"{p.as_posix()!r} is not a relative path.")
    return p


def _is_absolute_dir_path(p: Path) -> Path:
    if not isinstance(p, Path):
        # Let pydantic handle the type error.
        return p
    if p.suffix:
        raise ValueError(f"{p.as_posix()!r} is not a directory.")
    if not p.is_absolute():
        raise ValueError(f"{p.as_posix()!r} is not an absolute path.")
    return p


RelativeFilePath: TypeAlias = Annotated[Path, PlainValidator(_is_relative_file_path)]
AbsoluteFilePath: TypeAlias = Annotated[Path, PlainValidator(_is_absolute_file_path)]
RelativeDirPath: TypeAlias = Annotated[Path, PlainValidator(_is_relative_dir_path)]
AbsoluteDirPath: TypeAlias = Annotated[Path, PlainValidator(_is_absolute_dir_path)]


ValidationType: TypeAlias = Literal["dev", "prod"]



from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import DataSetId
from cognite_toolkit._cdf_tk.client.resource_classes.robotics._robot import RobotRequest, RobotResponse


class RobotsAPI(CDFResourceAPI[RobotResponse]):
    """API for managing Robot resources in CDF."""

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/robotics/robots", item_limit=1000, concurrency_max_workers=1),
                "retrieve": Endpoint(
                    method="POST", path="/robotics/robots/byids", item_limit=1000, concurrency_max_workers=1
                ),
                "update": Endpoint(
                    method="POST", path="/robotics/robots/update", item_limit=1000, concurrency_max_workers=1
                ),
                "delete": Endpoint(
                    method="POST", path="/robotics/robots/delete", item_limit=1000, concurrency_max_workers=1
                ),
                "list": Endpoint(method="GET", path="/robotics/robots", item_limit=1000),
            },
            disable_gzip=True,
        )

    def _validate_page_response(self, response: SuccessResponse | ItemsSuccessResponse) -> PagedResponse[RobotResponse]:
        return PagedResponse[RobotResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[DataSetId]:
        return ResponseItems[DataSetId].model_validate_json(response.body)

    def create(self, items: Sequence[RobotRequest]) -> list[RobotResponse]:
        """Create robots in CDF.

        Args:
            items: List of RobotRequest objects to create.
        Returns:
            List of created RobotResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(self, items: Sequence[DataSetId], ignore_unknown_ids: bool = False) -> list[RobotResponse]:
        """Retrieve robots from CDF.

        Args:
            items: List of DataSetId objects to retrieve.
            ignore_unknown_ids: If True, ignores unknown IDs during retrieval.
        Returns:
            List of retrieved RobotResponse objects.
        """
        if ignore_unknown_ids:
            return self._request_item_split_retries(items, method="retrieve")
        return self._request_item_response(items, method="retrieve")

    def update(
        self, items: Sequence[RobotRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[RobotResponse]:
        """Update robots in CDF.

        Args:
            items: List of RobotRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated RobotResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[DataSetId], ignore_unknown_ids: bool = False) -> None:
        """Delete robots from CDF.

        Args:
            items: List of DataSetId objects to delete.
            ignore_unknown_ids: If True, ignores unknown IDs during deletion.
        """
        if ignore_unknown_ids:
            self._request_item_split_retries_no_response(items, "delete")
        else:
            self._request_no_response(items, "delete")

    def paginate(
        self,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[RobotResponse]:
        """Paginate over robots in CDF.

        Args:
            limit: Maximum number of items per page.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of RobotResponse objects.
        """
        return self._paginate(cursor=cursor, limit=limit)

    def iterate(
        self,
        limit: int | None = 100,
    ) -> Iterable[list[RobotResponse]]:
        """Iterate over all robots in CDF.

        Args:
            limit: Maximum number of items to return. None returns all.

        Returns:
            Iterable of lists of RobotResponse objects.
        """
        return self._iterate(limit=limit)

    def list(self, limit: int | None = 100) -> list[RobotResponse]:
        """List all robots in CDF.

        Args:
            limit: Maximum number of items to return. None returns all.

        Returns:
            List of RobotResponse objects.
        """
        return self._list(limit=limit)



import builtins
from abc import ABC
from typing import Annotated, Any, Generic, Literal, TypeAlias

from pydantic import Field, JsonValue, TypeAdapter, field_serializer, field_validator

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
    T_RequestResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import TypedEdgeIdentifier, TypedNodeIdentifier

from ._references import ContainerReference, NodeReference, ViewReference


class InstanceDefinition(BaseModelObject, ABC):
    """Base class for node and edge instances."""

    instance_type: str  # "node" | "edge"
    space: str
    external_id: str


class InstanceSource(BaseModelObject):
    source: ViewReference | ContainerReference
    properties: dict[str, JsonValue] | None = None

    @field_serializer("source", mode="plain")
    def serialize_source(self, value: ViewReference | ContainerReference) -> Any:
        return {**value.dump(), "type": value.type}


class InstanceRequestDefinition(InstanceDefinition, RequestResource, ABC):
    existing_version: int | None = None
    sources: list[InstanceSource] | None = None


class InstanceResponseDefinition(InstanceDefinition, ResponseResource, Generic[T_RequestResource], ABC):
    version: int
    created_time: int
    last_updated_time: int
    deleted_time: int | None = None
    properties: dict[ViewReference | ContainerReference, dict[str, JsonValue]] | None = None

    @field_validator("properties", mode="before")
    def parse_reference(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        parsed: dict[ViewReference | ContainerReference, dict[str, Any]] = {}
        for space, inner_dict in value.items():
            if isinstance(space, ViewReference | ContainerReference):
                parsed[space] = inner_dict
                continue
            if not isinstance(inner_dict, dict) or not isinstance(space, str):
                raise ValueError(
                    f"Invalid properties format expected dict[str, dict[...]], got: dict[{type(space).__name__}, {type(inner_dict).__name__}]"
                )
            for view_or_container_identifier, prop in inner_dict.items():
                if not isinstance(view_or_container_identifier, str):
                    raise ValueError(
                        "Invalid properties format expected dict[str, dict[str, ...]]], "
                        f"got: dict[{type(space).__name__}, "
                        f"dict[{type(view_or_container_identifier).__name__}, ...]]"
                    )
                source_ref: ViewReference | ContainerReference
                if "/" in view_or_container_identifier:
                    external_id, version = view_or_container_identifier.split("/", 1)
                    source_ref = ViewReference(space=space, external_id=external_id, version=version)
                else:
                    source_ref = ContainerReference(space=space, external_id=view_or_container_identifier)
                parsed[source_ref] = prop
        return parsed

    @field_serializer("properties", mode="plain")
    def serialize_properties(self, value: dict[ViewReference | ContainerReference, dict[str, Any]] | None) -> Any:
        if value is None:
            return None
        serialized: dict[str, dict[str, Any]] = {}
        for source_ref, props in value.items():
            space = source_ref.space
            if space not in serialized:
                serialized[space] = {}
            if isinstance(source_ref, ViewReference):
                identifier = f"{source_ref.external_id}/{source_ref.version}"
            else:
                identifier = source_ref.external_id
            serialized[space][identifier] = props
        return serialized


class NodeRequest(InstanceRequestDefinition):
    """A node request resource."""

    instance_type: Literal["node"] = "node"
    type: NodeReference | None = None

    def as_id(self) -> TypedNodeIdentifier:
        return TypedNodeIdentifier(space=self.space, external_id=self.external_id)


class EdgeRequest(InstanceRequestDefinition):
    """An edge request resource."""

    instance_type: Literal["edge"] = "edge"
    type: NodeReference
    start_node: NodeReference
    end_node: NodeReference

    def as_id(self) -> TypedEdgeIdentifier:
        return TypedEdgeIdentifier(space=self.space, external_id=self.external_id)


class NodeResponse(InstanceResponseDefinition[NodeRequest]):
    """A node response from the API."""

    instance_type: Literal["node"] = "node"
    type: NodeReference | None = None

    def as_id(self) -> TypedNodeIdentifier:
        return TypedNodeIdentifier(space=self.space, external_id=self.external_id)

    @classmethod
    def request_cls(cls) -> builtins.type[NodeRequest]:
        return NodeRequest

    def as_request_resource(self) -> NodeRequest:
        dumped = self.dump()
        if self.properties:
            dumped["sources"] = [
                InstanceSource(source=source_ref, properties=props) for source_ref, props in self.properties.items()
            ]
        dumped["existingVersion"] = dumped.pop("version", None)
        return NodeRequest.model_validate(dumped, extra="ignore")


class EdgeResponse(InstanceResponseDefinition[EdgeRequest]):
    """An edge response from the API."""

    instance_type: Literal["edge"] = "edge"
    type: NodeReference
    start_node: NodeReference
    end_node: NodeReference

    def as_id(self) -> TypedEdgeIdentifier:
        return TypedEdgeIdentifier(space=self.space, external_id=self.external_id)

    @classmethod
    def request_cls(cls) -> builtins.type[EdgeRequest]:
        return EdgeRequest

    def as_request_resource(self) -> EdgeRequest:
        dumped = self.dump()
        if self.properties:
            dumped["sources"] = [
                InstanceSource(source=source_ref, properties=props) for source_ref, props in self.properties.items()
            ]
        dumped["existingVersion"] = dumped.pop("version", None)
        return EdgeRequest.model_validate(dumped, extra="ignore")


class InstanceSlimDefinition(BaseModelObject):
    """Slim version of instance definition for listing instances."""

    instance_type: Literal["node", "edge"]
    version: int
    was_modified: bool
    space: str
    external_id: str
    created_time: int
    last_updated_time: int

    def as_id(self) -> TypedNodeIdentifier | TypedEdgeIdentifier:
        if self.instance_type == "node":
            return TypedNodeIdentifier(space=self.space, external_id=self.external_id)
        else:
            return TypedEdgeIdentifier(space=self.space, external_id=self.external_id)


InstanceRequest: TypeAlias = Annotated[
    NodeRequest | EdgeRequest,
    Field(discriminator="instance_type"),
]
InstanceResponse: TypeAlias = Annotated[
    NodeResponse | EdgeResponse,
    Field(discriminator="instance_type"),
]

InstanceRequestAdapter: TypeAdapter[InstanceRequest] = TypeAdapter(InstanceRequest)



from collections.abc import Iterable, Mapping, Sequence
from types import MappingProxyType
from typing import Any, ClassVar, Literal, cast

from cognite.client.data_classes.aggregations import Count
from cognite.client.data_classes.data_modeling import EdgeId, NodeId, ViewId
from cognite.client.utils._identifier import InstanceId

from cognite_toolkit._cdf_tk import constants
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.request_classes.filters import InstanceFilter
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    ContainerReference,
    EdgeProperty,
    InstanceRequest,
    InstanceResponse,
    QueryEdgeExpression,
    QueryEdgeTableExpression,
    QueryNodeExpression,
    QueryNodeTableExpression,
    QueryRequest,
    QueryResponseTyped,
    QuerySelect,
    QuerySelectSource,
    QuerySortSpec,
    SpaceReference,
    ViewReference,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._instance import InstanceRequestAdapter
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import (
    TypedEdgeIdentifier,
    TypedNodeIdentifier,
    TypedViewReference,
)
from cognite_toolkit._cdf_tk.cruds import ContainerCRUD, SpaceCRUD, ViewCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.utils import sanitize_filename
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from . import StorageIOConfig
from ._base import ConfigurableStorageIO, Page, UploadableStorageIO
from .selectors import InstanceFileSelector, InstanceSelector, InstanceSpaceSelector, InstanceViewSelector, SelectedView


class InstanceIO(
    ConfigurableStorageIO[InstanceSelector, InstanceResponse],
    UploadableStorageIO[InstanceSelector, InstanceResponse, InstanceRequest],
):
    """This class provides functionality to interact with instances in Cognite Data Fusion (CDF).

    It is used to download, upload, and purge instances, as well as spaces,views, and containers related to instances.

    Args:
        client (ToolkitClient): An instance of ToolkitClient to interact with the CDF API.
        remove_existing_version (bool): Whether to remove existing versions of instances during upload.
            Default is True. Existing versions are used to safeguard against accidental overwrites.

    """

    KIND = "Instances"
    DISPLAY_NAME = "Instances"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson", ".yaml", ".yml"})
    CHUNK_SIZE = 1000
    UPLOAD_ENDPOINT = "/models/instances"
    UPLOAD_EXTRA_ARGS: ClassVar[Mapping[str, JsonVal] | None] = MappingProxyType({"autoCreateDirectRelations": True})
    BASE_SELECTOR = InstanceSelector

    def __init__(self, client: ToolkitClient, remove_existing_version: bool = True) -> None:
        super().__init__(client)
        self._remove_existing_version = remove_existing_version
        # Cache for view to read-only properties mapping
        self._view_readonly_properties_cache: dict[ViewReference, set[str]] = {}
        self._view_crud = ViewCRUD.create_loader(self.client)

    def as_id(self, item: InstanceResponse) -> str:
        return f"{item.space}:{item.external_id}"

    @staticmethod
    def _build_list_filter(selector: InstanceViewSelector | InstanceSpaceSelector) -> InstanceFilter:
        """Build an InstanceFilter from a selector.

        Args:
            selector: The selector to build the filter from.

        Returns:
            An InstanceFilter for the toolkit instances API.
        """
        source: TypedViewReference | None = None
        space: list[str] | None = None

        if isinstance(selector, InstanceViewSelector):
            source = TypedViewReference(
                space=selector.view.space,
                external_id=selector.view.external_id,
                version=selector.view.version or "",
            )
            if selector.instance_spaces:
                space = list(selector.instance_spaces)
        elif isinstance(selector, InstanceSpaceSelector):
            space = [selector.instance_space]
            if selector.view and selector.view.version:
                source = TypedViewReference(
                    space=selector.view.space,
                    external_id=selector.view.external_id,
                    version=selector.view.version,
                )

        return InstanceFilter(
            instance_type=selector.instance_type,
            source=source,
            space=space,
        )

    @staticmethod
    def _build_query_filter(selector: InstanceViewSelector, instance_type: Literal["node", "edge"]) -> dict[str, Any]:
        """Build a filter dict for the query endpoint from an InstanceViewSelector."""
        leaf_filters: list[dict[str, Any]] = [
            {"hasData": [{**selector.view.as_id().dump(), "type": "view"}]},
        ]
        if selector.instance_spaces and len(selector.instance_spaces) == 1:
            leaf_filters.append(
                {"equals": {"property": [instance_type, "space"], "value": selector.instance_spaces[0]}}
            )
        elif selector.instance_spaces and len(selector.instance_spaces) > 1:
            leaf_filters.append(
                {"in": {"property": [instance_type, "space"], "values": list(selector.instance_spaces)}}
            )
        if len(leaf_filters) == 1:
            return leaf_filters[0]
        return {"and": leaf_filters}

    def _filter_readonly_properties(self, instance: InstanceRequest) -> None:
        """Filter out read-only properties from the instance.

        Warnings: This mutates the instance in-place.

        This is as of 17/02/26, the path, root, and pathLastUpdatedTime time in the CogniteAsset container,
        and isUploaded and uploadedTime in the CogniteFile container.
        """
        if not instance.sources:
            return

        for source in instance.sources:
            if source.properties is None:
                continue
            readonly_properties: set[str] = set()
            if isinstance(source.source, ViewReference):
                if source.source not in self._view_readonly_properties_cache:
                    self._view_readonly_properties_cache[source.source] = self._view_crud.get_readonly_properties(
                        source.source
                    )
                readonly_properties = self._view_readonly_properties_cache[source.source]
            elif isinstance(source.source, ContainerReference):
                if source.source.as_tuple() in constants.READONLY_CONTAINER_PROPERTIES:
                    readonly_properties = constants.READONLY_CONTAINER_PROPERTIES[source.source.as_tuple()]
            if not readonly_properties:
                continue
            source.properties = {k: v for k, v in source.properties.items() if k not in readonly_properties}

    def stream_data(self, selector: InstanceSelector, limit: int | None = None) -> Iterable[Page]:
        if isinstance(selector, InstanceViewSelector) and selector.include_edges and selector.instance_type == "node":
            yield from self._instances_with_container_and_edge_properties(selector, limit)
        elif isinstance(selector, InstanceViewSelector | InstanceSpaceSelector):
            yield from self._instances_with_container_properties(selector, limit)
        elif isinstance(selector, InstanceFileSelector):
            for chunk in chunker_sequence(selector.ids, self.CHUNK_SIZE):
                yield Page(worker_id="main", items=self.client.tool.instances.retrieve(chunk))
        else:
            raise NotImplementedError()

    def _instances_with_container_and_edge_properties(
        self, selector: InstanceViewSelector, limit: int | None
    ) -> Iterable[Page]:
        instance_filter = self._build_query_filter(selector, "node")
        views = self.client.tool.views.retrieve([selector.view.as_id()])
        if not views:
            raise ToolkitValueError(f"The view {selector.view.as_id()} does not exist")
        view = views[0]
        with_: dict[str, QueryNodeExpression | QueryEdgeExpression] = {
            "nodes": QueryNodeExpression(
                limit=min(self.CHUNK_SIZE, limit) if limit is not None else self.CHUNK_SIZE,
                nodes=QueryNodeTableExpression(filter=instance_filter),
                # Sort to ensure performance. f you do not sort, you get the internal index,
                # which includes all deleted instances as well.
                sort=[QuerySortSpec(property=["node", "space"]), QuerySortSpec(property=["node", "externalId"])],
            )
        }
        select: dict[str, QuerySelect] = {
            "nodes": QuerySelect(sources=[QuerySelectSource(source=view.as_id(), properties=["*"])])
        }
        edge_ids: list[str] = []
        for prop_id, prop in view.properties.items():
            if not isinstance(prop, EdgeProperty):
                continue
            with_[prop_id] = QueryEdgeExpression(
                edges=QueryEdgeTableExpression(
                    from_="nodes",
                    chain_to="source" if prop.direction == "outwards" else "destination",
                    direction=prop.direction,
                ),
                sort=[QuerySortSpec(property=["edge", "space"]), QuerySortSpec(property=["edge", "externalId"])],
            )
            edge_ids.append(prop_id)
            select[prop_id] = QuerySelect()

        query = QueryRequest(with_=with_, select=select)
        total = 0
        while True:
            response = self._exhaust_edge_queries(query, edge_ids)
            nodes = response.items.get("nodes", [])
            # De-duplicate edges across properties, as the same edge can be returned for multiple
            # properties if it connects two nodes that are in the result set.
            edges: dict[TypedNodeIdentifier | TypedEdgeIdentifier, InstanceResponse] = {}
            for prop_id in edge_ids:
                for edge in response.items.get(prop_id, []):
                    ref = edge.as_id()
                    if ref not in edges:
                        edges[ref] = edge
            items = nodes + list(edges.values())
            total += len(nodes)
            next_cursor = response.next_cursor.get("nodes")
            yield Page(worker_id="main", items=items, next_cursor=next_cursor)
            if next_cursor is None or (limit is not None and total >= limit) or not nodes:
                break
            page_limit = min(self.CHUNK_SIZE, limit - total) if limit is not None else self.CHUNK_SIZE
            query.with_["nodes"].limit = page_limit
            query.cursors = {"nodes": next_cursor}

    def _exhaust_edge_queries(self, query: QueryRequest, edge_properties: list[str]) -> QueryResponseTyped:
        """Exhausts the edge queries in the with_ clause of the query until all cursors are None.

        This is necessary to ensure that we get all edges for the nodes in the result set, as edges can be returned
        on multiple properties if they connect two nodes that are in the result set.

        Args:
            query: The query to execute.
            edge_properties: The list of edge properties to exhaust.

        Returns:
            The final QueryResponse with all edge queries exhausted.
        """
        first: QueryResponseTyped | None = None
        while True:
            response = self.client.tool.instances.query(query, type_results=True)
            if first is None:
                first = response
            else:
                for key, items in response.items.items():
                    if key not in first.items:
                        # In practice, this is unreachable. It is just in case.
                        first.items[key] = items
                    else:
                        first.items[key].extend(items)
            next_cursors: dict[str, str] = {}
            for prop_id in edge_properties:
                edge_cursor = response.next_cursor.get(prop_id)
                if edge_cursor is not None:
                    next_cursors[prop_id] = edge_cursor
            if not next_cursors or not response.items:
                return first
            node_cursor = response.next_cursor.get("nodes")
            if node_cursor is not None:
                next_cursors["nodes"] = node_cursor
            query = query.model_copy(update={"cursors": next_cursors})

    def _instances_with_container_properties(
        self, selector: InstanceViewSelector | InstanceSpaceSelector, limit: int | None
    ) -> Iterable[Page]:
        instance_filter = self._build_list_filter(selector)
        total = 0
        cursor: str | None = None
        while cursor is not None or total == 0:
            page_limit = min(self.CHUNK_SIZE, limit - total) if limit is not None else self.CHUNK_SIZE
            page = self.client.tool.instances.paginate(instance_filter, limit=page_limit, cursor=cursor)
            total += len(page.items)
            if page:
                yield Page(worker_id="main", items=page.items, next_cursor=page.next_cursor)
            if page.next_cursor is None or (limit is not None and total >= limit) or not page.items:
                break
            cursor = page.next_cursor

    def download_ids(self, selector: InstanceSelector, limit: int | None = None) -> Iterable[Sequence[InstanceId]]:
        # Todo: Switch to use pydantic classes once purge has been updated.
        if isinstance(selector, InstanceFileSelector) and selector.validate_instance is False:
            instances_to_yield = selector.instance_ids
            if limit is not None:
                instances_to_yield = instances_to_yield[:limit]
            yield from chunker_sequence(instances_to_yield, self.CHUNK_SIZE)
        else:
            yield from (
                [
                    NodeId(space=instance.space, external_id=instance.external_id)
                    if instance.instance_type == "node"
                    else EdgeId(space=instance.space, external_id=instance.external_id)
                    for instance in chunk.items
                ]
                for chunk in self.stream_data(selector, limit)
            )

    def count(self, selector: InstanceSelector) -> int | None:
        if isinstance(selector, InstanceViewSelector) or (
            isinstance(selector, InstanceSpaceSelector) and selector.view
        ):
            view_id = cast(SelectedView, selector.view)
            result = self.client.data_modeling.instances.aggregate(
                view=ViewId(space=view_id.space, external_id=view_id.external_id, version=view_id.version),
                aggregates=Count("externalId"),
                instance_type=selector.instance_type,
                space=selector.get_instance_spaces(),
            )
            return int(result.value or 0)
        elif isinstance(selector, InstanceSpaceSelector):
            statistics = self.client.data_modeling.statistics.spaces.retrieve(space=selector.instance_space)
            if statistics is None:
                return None
            if selector.instance_type == "node":
                return statistics.nodes
            elif selector.instance_type == "edge":
                return statistics.edges
            # This should never happen due to validation in the selector.
            raise ValueError(f"Unknown instance type {selector.instance_type!r}")
        elif isinstance(selector, InstanceFileSelector):
            return len(selector.items)
        raise NotImplementedError()

    def data_to_json_chunk(
        self, data_chunk: Sequence[InstanceResponse], selector: InstanceSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        return [instance.as_request_resource().dump() for instance in data_chunk]

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> InstanceRequest:
        item_to_load = item_json.copy()
        if self._remove_existing_version and "existingVersion" in item_to_load:
            del item_to_load["existingVersion"]
        instance = InstanceRequestAdapter.validate_python(item_to_load)
        self._filter_readonly_properties(instance)
        return instance

    def configurations(self, selector: InstanceSelector) -> Iterable[StorageIOConfig]:
        if not isinstance(selector, InstanceViewSelector | InstanceSpaceSelector):
            return
        spaces = list(set((selector.get_instance_spaces() or []) + (selector.get_schema_spaces() or [])))
        if not spaces:
            return
        space_crud = SpaceCRUD.create_loader(self.client)
        retrieved_spaces = space_crud.retrieve([SpaceReference(space=space) for space in spaces])
        retrieved_spaces = [space for space in retrieved_spaces if not space.is_global]
        if not retrieved_spaces:
            return
        for space in retrieved_spaces:
            yield StorageIOConfig(
                kind=SpaceCRUD.kind,
                folder_name=SpaceCRUD.folder_name,
                value=space_crud.dump_resource(space),
                filename=sanitize_filename(space.space),
            )
        if not selector.view:
            return
        view_crud = ViewCRUD(self.client, None, None, topological_sort_implements=True)
        views = self.client.tool.views.retrieve([selector.view.as_id()], include_inherited_properties=False)
        views = [view for view in views if not view.is_global]
        if not views:
            return
        for view in views:
            filename = f"{view.space}_{view.external_id}"
            if view.version is not None:
                filename += f"_{view.version}"
            yield StorageIOConfig(
                kind=ViewCRUD.kind,
                folder_name=ViewCRUD.folder_name,
                value=view_crud.dump_resource(view),
                filename=sanitize_filename(filename),
            )
        container_ids = list({container for view in views for container in view.mapped_containers})
        if not container_ids:
            return
        container_crud = ContainerCRUD.create_loader(self.client)
        containers = container_crud.retrieve(container_ids)
        containers = [container for container in containers if not container.is_global]
        if not containers:
            return
        for container in containers:
            yield StorageIOConfig(
                kind=ContainerCRUD.kind,
                folder_name=ContainerCRUD.folder_name,
                value=container_crud.dump_resource(container),
                filename=sanitize_filename(f"{container.space}_{container.external_id}"),
            )



from typing import Any, Literal, TypeAlias

from pydantic import Field, JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, ResponseResource, UpdatableRequestResource

from .identifiers import ExternalId, InternalId

AnnotationStatus: TypeAlias = Literal["suggested", "rejected", "approved"]
AnnotationType: TypeAlias = Literal[
    "diagrams.AssetLink",
    "diagrams.FileLink",
    "diagrams.InstanceLink",
    "diagrams.Junction",
    "diagrams.Line",
    "diagrams.UnhandledSymbolObject",
    "diagrams.UnhandledTextObject",
    "documents.ExtractedText",
    "forms.Detection",
    "images.AssetLink",
    "images.Classification",
    "images.InstanceLink",
    "images.KeypointCollection",
    "images.ObjectDetection",
    "images.TextRegion",
    "isoplan.IsoPlanAnnotation",
    "pointcloud.BoundingVolume",
]


class BoundingBox(BaseModelObject):
    confidence: float | None = None
    x_min: float
    x_max: float
    y_min: float
    y_max: float


class AssetLinkData(BaseModelObject):
    asset_ref: InternalId | ExternalId
    text_region: BoundingBox
    symbol_region: BoundingBox | None = None
    page_number: int | None = None
    text: str | None = None
    description: str | None = None


class FileLinkData(BaseModelObject):
    file_ref: InternalId | ExternalId
    text_region: BoundingBox
    symbol_region: BoundingBox | None = None
    page_number: int | None = None
    text: str | None = None
    description: str | None = None


class Annotation(BaseModelObject):
    annotated_resource_type: str
    annotated_resource_id: int
    annotation_type: AnnotationType
    creating_app: str
    creating_app_version: str
    creating_user: str | None
    data: AssetLinkData | FileLinkData | dict[str, JsonValue]
    status: AnnotationStatus


class AnnotationRequest(Annotation, UpdatableRequestResource):
    """Request data class for annotations."""

    # The 'id' field is not part of the request when creating a new resource,
    # but is needed when updating an existing resource.
    id: int | None = Field(default=None, exclude=True)

    def as_id(self) -> InternalId:
        if self.id is None:
            raise ValueError("Cannot convert AnnotationRequest to InternalId when id is None")
        return InternalId(id=self.id)

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        """Converts the request to an update payload for the API."""
        if self.id is None:
            raise ValueError("id must be provided to create an update dictionary")
        return {
            "id": self.id,
            "update": {
                "annotationType": {"set": self.annotation_type},
                "data": {"set": self.data.dump() if isinstance(self.data, AssetLinkData | FileLinkData) else self.data},
                "status": {"set": self.status},
            },
        }


class AnnotationResponse(Annotation, ResponseResource[AnnotationRequest]):
    """Response data class for annotations."""

    id: int
    created_time: int
    last_updated_time: int

    def as_id(self) -> InternalId:
        return InternalId(id=self.id)

    @classmethod
    def request_cls(cls) -> type[AnnotationRequest]:
        return AnnotationRequest



from typing import Literal

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource

from .identifiers import ExternalId


class SimulatorModelRevision(BaseModelObject):
    external_id: str
    model_external_id: str
    description: str | None = None
    file_id: int


class SimulatorModelRevisionRequest(RequestResource, SimulatorModelRevision):
    """Request class for creating a simulator model revision."""

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class SimulatorModelRevisionResponse(ResponseResource[SimulatorModelRevisionRequest], SimulatorModelRevision):
    """Response class for a simulator model revision."""

    id: int
    description: str | None = None
    simulator_external_id: str
    created_by_user_id: str
    status: Literal["unknown", "success", "failure"] = "unknown"
    status_message: str | None = None
    version_number: int
    log_id: int
    data_set_id: int
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[SimulatorModelRevisionRequest]:
        return SimulatorModelRevisionRequest



"""This module contains data classes for the APM Config node used to configure Infield.

These classes are used to represent the configuration of the APM Config node in a structured way, such that it can be
used in the InfieldV1 Loader and thus be represented as a resource type in Toolkit. We do not do any validation in the
FeatureConfiguration objects as this is just JSON object in the node, but use the structure to do lookup of
data sets, spaces, and groups.
"""

from typing import ClassVar, Literal

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import (
    TypedNodeIdentifier,
    TypedViewReference,
    WrappedInstanceRequest,
    WrappedInstanceResponse,
)

APM_CONFIG_SPACE: Literal["APM_Config"] = "APM_Config"
APM_CONFIG_VIEW_ID = TypedViewReference(space=APM_CONFIG_SPACE, external_id="APM_Config", version="1")


class EnabledToggle(BaseModelObject):
    enabled: bool | None = None


class DisabledToggle(BaseModelObject):
    disabled: bool | None = None


class ThreeDModelIdentifier(BaseModelObject):
    revision_id: int | None = None
    model_id: int | None = None
    name: str | None = None


class ThreeDConfiguration(BaseModelObject):
    full_weight_models: list[ThreeDModelIdentifier] | None = None
    light_weight_models: list[ThreeDModelIdentifier] | None = None


class ResourceFilters(BaseModelObject):
    data_set_ids: list[int] | None = None
    asset_subtree_external_ids: list[str] | None = None
    root_asset_external_ids: list[str] | None = None
    external_id_prefix: str | None = None
    spaces: list[str] | None = None


class RootLocationDataFilters(BaseModelObject):
    general: ResourceFilters | None = None
    assets: ResourceFilters | None = None
    files: ResourceFilters | None = None
    timeseries: ResourceFilters | None = None


class ObservationFeatureToggles(BaseModelObject):
    is_enabled: bool | None = None
    is_write_back_enabled: bool | None = None
    notifications_endpoint_external_id: str | None = None
    attachments_endpoint_external_id: str | None = None


class RootLocationFeatureToggles(BaseModelObject):
    three_d: bool | None = None
    trends: bool | None = None
    documents: bool | None = None
    workorders: bool | None = None
    notifications: bool | None = None
    media: bool | None = None
    template_checklist_flow: bool | None = None
    workorder_checklist_flow: bool | None = None
    observations: ObservationFeatureToggles | None = None


class ObservationConfigFieldProperty(BaseModelObject):
    display_title: str | None = None
    display_description: str | None = None
    is_required: bool | None = None


class ObservationConfigDropdownPropertyOption(BaseModelObject):
    id: str | None = None
    value: str | None = None
    label: str | None = None


class ObservationConfigDropdownProperty(ObservationConfigFieldProperty):
    options: list[ObservationConfigDropdownPropertyOption] | None = None


class ObservationsConfig(BaseModelObject):
    files: ObservationConfigFieldProperty | None = None
    description: ObservationConfigFieldProperty | None = None
    asset: ObservationConfigFieldProperty | None = None
    troubleshooting: ObservationConfigFieldProperty | None = None
    type: ObservationConfigDropdownProperty | None = None
    priority: ObservationConfigDropdownProperty | None = None


class RootLocationConfiguration(BaseModelObject):
    asset_external_id: str | None = None
    external_id: str | None = None
    display_name: str | None = None
    three_d_configuration: ThreeDConfiguration | None = None
    data_set_id: int | None = None
    template_admins: list[str] | None = None  # list of Group Names
    checklist_admins: list[str] | None = None  # list of Group Names
    app_data_instance_space: str | None = None
    source_data_instance_space: str | None = None
    data_filters: RootLocationDataFilters | None = None
    feature_toggles: RootLocationFeatureToggles | None = None
    observations: ObservationsConfig | None = None


class Activities(BaseModelObject):
    overview_card: dict[str, JsonValue] | None = None


class Documents(BaseModelObject):
    type: str | None = None
    title: str | None = None
    description: str | None = None


class Discipline(BaseModelObject):
    name: str | None = None
    external_id: str | None = None


class Notification(BaseModelObject):
    overview_card: dict[str, JsonValue] | None = None


class CabinetConfiguration(BaseModelObject):
    enable_plan_analysis: bool | None = None
    enable_hourly_optimisation: bool | None = None
    enable_resource_allocation: bool | None = None


class AssetPagePropertyCardConfiguration(BaseModelObject):
    linkable_asset_keys: list[str] | None = None
    highlighted_properties: list[str] | None = None


class AssetPageConfiguration(BaseModelObject):
    property_card: AssetPagePropertyCardConfiguration


class FeatureConfiguration(BaseModelObject):
    root_location_configurations: list[RootLocationConfiguration] | None = None
    copilot: EnabledToggle | DisabledToggle | None = None
    activities: Activities | None = None
    documents: Documents | None = None
    disciplines: list[Discipline] | None = None
    notifications: Notification | None = None
    observations: JsonValue | None = None
    psn_configuration: EnabledToggle | DisabledToggle | None = None
    canvas_configuration: EnabledToggle | DisabledToggle | None = None
    cabinet_configuration: CabinetConfiguration | None = None
    asset_page_configuration: AssetPageConfiguration | None = None
    subactivities_configuration: EnabledToggle | DisabledToggle | None = None


class APMConfig(BaseModelObject):
    """Base class for APM Config containing common properties."""

    name: str | None = None
    app_data_space_id: str | None = None
    app_data_space_version: str | None = None
    customer_data_space_id: str | None = None
    customer_data_space_version: str | None = None
    feature_configuration: FeatureConfiguration | None = None


class APMConfigRequest(WrappedInstanceRequest, APMConfig):
    """APM Config request resource for creating/updating nodes."""

    VIEW_ID: ClassVar[TypedViewReference] = APM_CONFIG_VIEW_ID
    instance_type: Literal["node"] = "node"
    space: Literal["APM_Config"] = APM_CONFIG_SPACE

    def as_id(self) -> TypedNodeIdentifier:
        return TypedNodeIdentifier(space=self.space, external_id=self.external_id)


class APMConfigResponse(WrappedInstanceResponse[APMConfigRequest], APMConfig):
    """APM Config response resource returned from API."""

    VIEW_ID: ClassVar[TypedViewReference] = APM_CONFIG_VIEW_ID
    instance_type: Literal["node"] = "node"
    space: Literal["APM_Config"] = APM_CONFIG_SPACE

    @classmethod
    def request_cls(cls) -> type[APMConfigRequest]:
        return APMConfigRequest

    def as_request_resource(self) -> APMConfigRequest:
        # InstanceType and space are constants, so we exclude them.
        return APMConfigRequest.model_validate(
            self.model_dump(mode="json", by_alias=True, exclude_unset=True, exclude={"instance_type", "space"}),
            extra="ignore",
        )



from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId

from .base import ToolkitResource


class WorkflowYAML(ToolkitResource):
    external_id: str = Field(
        max_length=255,
        description="Identifier for a workflow. Must be unique for the project."
        " No trailing or leading whitespace and no null characters allowed.",
    )
    description: str | None = Field(None, max_length=500)
    data_set_external_id: str | None = None
    max_concurrent_executions: int | None = Field(
        default=None,
        description="Maximum concurrent executions for this workflow. Defaults to the project limit if not specified or explicitly set to None. Values exceeding the project limit are dynamically capped at runtime.",
    )

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)



import builtins
from typing import ClassVar, Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
)

from ._auth import (
    BasicAuthenticationRequest,
    BasicAuthenticationResponse,
    ClientCredentialAuthenticationRequest,
    ClientCredentialAuthenticationResponse,
    ScramShaAuthenticationRequest,
    ScramShaAuthenticationResponse,
)
from ._base import SourceRequestDefinition, SourceResponseDefinition
from ._certificate import AuthCertificateRequest, CACertificateRequest, CertificateResponse


class KafkaBroker(BaseModelObject):
    host: str
    port: int


class KafkaSource(BaseModelObject):
    type: Literal["kafka"] = "kafka"
    bootstrap_brokers: list[KafkaBroker]
    use_tls: bool | None = None


class KafkaSourceRequest(KafkaSource, SourceRequestDefinition):
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"use_tls"})
    authentication: (
        BasicAuthenticationRequest | ClientCredentialAuthenticationRequest | ScramShaAuthenticationRequest | None
    ) = Field(None, discriminator="type")
    ca_certificate: CACertificateRequest | None = None
    auth_certificate: AuthCertificateRequest | None = None


class KafkaSourceResponse(
    SourceResponseDefinition,
    KafkaSource,
    ResponseResource[KafkaSourceRequest],
):
    authentication: (
        BasicAuthenticationResponse | ClientCredentialAuthenticationResponse | ScramShaAuthenticationResponse | None
    ) = Field(None, discriminator="type")
    ca_certificate: CertificateResponse | None = None
    auth_certificate: CertificateResponse | None = None

    @classmethod
    def request_cls(cls) -> builtins.type[KafkaSourceRequest]:
        return KafkaSourceRequest



import builtins
from typing import ClassVar, Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
)

from ._auth import (
    BasicAuthenticationRequest,
    BasicAuthenticationResponse,
    ClientCredentialAuthenticationRequest,
    ClientCredentialAuthenticationResponse,
    HTTPBasicAuthenticationRequest,
    HTTPBasicAuthenticationResponse,
)
from ._base import SourceRequestDefinition, SourceResponseDefinition
from ._certificate import AuthCertificateRequest, CACertificateRequest, CertificateResponse


class RESTSource(BaseModelObject):
    type: Literal["rest"] = "rest"
    host: str
    port: int | None = None


class RESTSourceRequest(RESTSource, SourceRequestDefinition):
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"scheme", "port"})
    scheme: Literal["https", "http"] | None = None
    authentication: (
        BasicAuthenticationRequest | HTTPBasicAuthenticationRequest | ClientCredentialAuthenticationRequest | None
    ) = Field(None, discriminator="type")
    ca_certificate: CACertificateRequest | None = None
    auth_certificate: AuthCertificateRequest | None = None


class RESTSourceResponse(
    SourceResponseDefinition,
    RESTSource,
    ResponseResource[RESTSourceRequest],
):
    authentication: (
        BasicAuthenticationResponse | HTTPBasicAuthenticationResponse | ClientCredentialAuthenticationResponse | None
    ) = Field(None, discriminator="type")
    ca_certificate: CertificateResponse | None = None
    auth_certificate: CertificateResponse | None = None

    @classmethod
    def request_cls(cls) -> builtins.type[RESTSourceRequest]:
        return RESTSourceRequest



import builtins
from typing import Literal

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource
from cognite_toolkit._cdf_tk.constants import StreamTemplateName

from .identifiers import ExternalId


class Stream(BaseModelObject):
    external_id: str


class StreamRequest(Stream, RequestResource):
    """Stream request resource class."""

    settings: dict[Literal["template"], dict[Literal["name"], StreamTemplateName]]

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class LifecycleObject(BaseModelObject):
    """Lifecycle object."""

    data_deleted_after: str | None = None
    retained_after_soft_delete: str


class ResourceUsage(BaseModelObject):
    """Resource quota with provisioned and consumed values."""

    provisioned: int
    consumed: int | None = None


class LimitsObject(BaseModelObject):
    """Limits object."""

    max_records_total: ResourceUsage
    max_giga_bytes_total: ResourceUsage
    max_filtering_interval: str | None = None


class StreamSettings(BaseModelObject):
    """Stream settings object."""

    lifecycle: LifecycleObject
    limits: LimitsObject


class StreamResponse(Stream, ResponseResource[StreamRequest]):
    """Stream response resource class."""

    created_time: int
    created_from_template: StreamTemplateName
    type: Literal["Mutable", "Immutable"]
    settings: StreamSettings | None = None

    @classmethod
    def request_cls(cls) -> builtins.type[StreamRequest]:
        return StreamRequest

    def as_request_resource(self) -> StreamRequest:
        return StreamRequest.model_validate(
            {
                "externalId": self.external_id,
                "settings": {"template": {"name": self.created_from_template}},
            }
        )

    def as_write(self) -> StreamRequest:
        return StreamRequest.model_validate(
            {
                "externalId": self.external_id,
                "settings": {"template": {"name": self.created_from_template}},
            }
        )



from typing import Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource

from .data_modeling import DataModelReference
from .identifiers import ExternalId, InternalId


class LocationFilterScene(BaseModelObject):
    """Scene configuration for a location filter."""

    external_id: str
    space: str


class LocationFilterView(BaseModelObject):
    """View identifier with entity type for location filter."""

    external_id: str
    space: str
    version: str
    represents_entity: Literal["MAINTENANCE_ORDER", "OPERATION", "NOTIFICATION", "ASSET"] | None = None


class AssetCentricSubFilter(BaseModelObject):
    """Sub-filter for asset-centric resource types."""

    data_set_ids: list[int] | None = None
    asset_subtree_ids: list[ExternalId | InternalId] | None = None
    external_id_prefix: str | None = None


class AssetCentricFilter(BaseModelObject):
    """Filter definition for asset-centric resource types."""

    assets: AssetCentricSubFilter | None = None
    events: AssetCentricSubFilter | None = None
    files: AssetCentricSubFilter | None = None
    timeseries: AssetCentricSubFilter | None = None
    sequences: AssetCentricSubFilter | None = None
    data_set_ids: list[int] | None = None
    asset_subtree_ids: list[ExternalId | InternalId] | None = None
    external_id_prefix: str | None = None


class LocationFilter(BaseModelObject):
    """Base class for location filter with common fields."""

    external_id: str
    name: str
    description: str | None = None
    parent_id: int | None = None
    data_models: list[DataModelReference] | None = None
    instance_spaces: list[str] | None = None
    scene: LocationFilterScene | None = None
    asset_centric: AssetCentricFilter | None = None
    views: list[LocationFilterView] | None = None
    data_modeling_type: Literal["HYBRID", "DATA_MODELING_ONLY", "ASSET_CENTRIC_ONLY"] | None = None


class LocationFilterRequest(LocationFilter, RequestResource):
    """Request resource for creating/updating location filters."""

    # This is not part of the request payload, but we need it to identify existing resources for updates.
    id: int | None = Field(default=None, exclude=True)
    # Used for topological sorting and deferred parent ID lookup during creation.
    parent_external_id: str | None = Field(default=None, exclude=True)

    def as_id(self) -> InternalId:
        if self.id is None:
            raise ValueError("Cannot get ID for LocationFilterRequest without 'id' set.")
        return InternalId(id=self.id)


class LocationFilterResponse(LocationFilter, ResponseResource[LocationFilterRequest]):
    """Response resource for location filters."""

    id: int
    created_time: int
    last_updated_time: int
    locations: list["LocationFilterResponse"] | None = None

    @classmethod
    def request_cls(cls) -> type[LocationFilterRequest]:
        return LocationFilterRequest



from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.extraction_pipeline_config import (
    ExtractionPipelineConfigRequest,
    ExtractionPipelineConfigResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExtractionPipelineConfigId


class ExtractionPipelineConfigsAPI(CDFResourceAPI[ExtractionPipelineConfigResponse]):
    """API for managing extraction pipeline configuration revisions."""

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/extpipes/config", item_limit=1),
                "retrieve": Endpoint(method="GET", path="/extpipes/config", item_limit=1),
                "list": Endpoint(method="GET", path="/extpipes/config/revisions", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[ExtractionPipelineConfigResponse]:
        return PagedResponse[ExtractionPipelineConfigResponse].model_validate_json(response.body)

    def create(self, items: Sequence[ExtractionPipelineConfigRequest]) -> list[ExtractionPipelineConfigResponse]:
        """Create new configuration revisions for extraction pipelines.

        The CDF API creates one revision at a time, so items are created sequentially.

        Args:
            items: List of configuration revisions to create.

        Returns:
            List of created configuration revisions.
        """
        endpoint = self._method_endpoint_map["create"]
        results: list[ExtractionPipelineConfigResponse] = []
        for item in items:
            request = RequestMessage(
                endpoint_url=self._make_url(endpoint.path),
                method=endpoint.method,
                body_content=item.dump(),
            )
            response = self._http_client.request_single_retries(request).get_success_or_raise()
            results.append(ExtractionPipelineConfigResponse.model_validate_json(response.body))
        return results

    def retrieve(self, items: Sequence[ExtractionPipelineConfigId]) -> list[ExtractionPipelineConfigResponse]:
        """Retrieve configuration revisions by their identifiers.

        Each identifier specifies an extraction pipeline external ID and a revision number.

        Args:
            items: List of ExtractionPipelineConfigId objects to retrieve.

        Returns:
            List of retrieved configuration revisions.
        """
        endpoint = self._method_endpoint_map["retrieve"]
        results: list[ExtractionPipelineConfigResponse] = []
        for item in items:
            request = RequestMessage(
                endpoint_url=self._make_url(endpoint.path),
                method=endpoint.method,
                parameters=item.dump(),
            )
            response = self._http_client.request_single_retries(request).get_success_or_raise()
            results.append(ExtractionPipelineConfigResponse.model_validate_json(response.body))
        return results

    def paginate(
        self,
        external_id: str,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[ExtractionPipelineConfigResponse]:
        """Retrieve a page of configuration revisions for an extraction pipeline.

        Args:
            external_id: External ID of the extraction pipeline.
            limit: Maximum number of items to return per page.
            cursor: Cursor for pagination.

        Returns:
            A paged response containing configuration revisions.
        """
        return self._paginate(
            limit=limit,
            cursor=cursor,
            params={"externalId": external_id},
        )

    def iterate(
        self,
        external_id: str,
        limit: int | None = None,
    ) -> Iterable[list[ExtractionPipelineConfigResponse]]:
        """Iterate over all configuration revisions for an extraction pipeline.

        Args:
            external_id: External ID of the extraction pipeline.
            limit: Maximum number of items to return. None for all items.

        Yields:
            Lists of configuration revisions per page.
        """
        return self._iterate(
            limit=limit,
            params={"externalId": external_id},
        )

    def list(
        self,
        external_id: str,
        limit: int | None = None,
    ) -> list[ExtractionPipelineConfigResponse]:
        """List all configuration revisions for an extraction pipeline.

        Args:
            external_id: External ID of the extraction pipeline.
            limit: Maximum number of items to return. None for all items.

        Returns:
            List of configuration revisions.
        """
        return self._list(
            limit=limit,
            params={"externalId": external_id},
        )



from typing import Any, ClassVar, Literal

from pydantic import JsonValue, model_validator

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import DataModelReference
from cognite_toolkit._cdf_tk.utils.text import sanitize_instance_external_id

from .instance_api import (
    TypedInstanceIdentifier,
    TypedNodeIdentifier,
    TypedViewReference,
    WrappedInstanceListRequest,
    WrappedInstanceListResponse,
    WrappedInstanceRequest,
    WrappedInstanceResponse,
    move_properties,
)

INFIELD_LOCATION_CONFIG_VIEW_ID = TypedViewReference(
    space="cdf_infield", external_id="InFieldLocationConfig", version="v1"
)
INFIELD_ON_CDM_DATA_MODEL = DataModelReference(
    space="infield_cdm_source_desc_sche_asset_file_ts",
    external_id="InFieldOnCDM",
    version="v1",
)
INFIELD_CDM_LOCATION_CONFIG_VIEW_ID = TypedViewReference(
    space="infield_cdm_source_desc_sche_asset_file_ts", external_id="InFieldCDMLocationConfig", version="v1"
)
DATA_EXPLORATION_CONFIG_VIEW_ID = TypedViewReference(
    space="cdf_infield", external_id="DataExplorationConfig", version="v1"
)


class DataExplorationConfig(BaseModelObject):
    """Data Exploration Configuration resource class."""

    VIEW_ID: ClassVar[TypedViewReference] = DATA_EXPLORATION_CONFIG_VIEW_ID
    space: str | None = None
    external_id: str | None = None

    observations: dict[str, JsonValue] | None = None
    activities: dict[str, JsonValue] | None = None
    documents: dict[str, JsonValue] | None = None
    notifications: dict[str, JsonValue] | None = None
    assets: dict[str, JsonValue] | None = None

    @model_validator(mode="before")
    @classmethod
    def move_properties(cls, data: dict[str, Any]) -> dict[str, Any]:
        return move_properties(data, cls.VIEW_ID)


class InFieldLocationConfig(BaseModelObject):
    """Infield Location Configuration resource class.

    This class is used for both the response and request resource for Infield Location Configuration nodes.
    """

    VIEW_ID: ClassVar[TypedViewReference] = INFIELD_LOCATION_CONFIG_VIEW_ID
    root_location_external_id: str | None = None
    feature_toggles: dict[str, JsonValue] | None = None
    app_instance_space: str | None = None
    access_management: dict[str, JsonValue] | None = None
    data_filters: dict[str, JsonValue] | None = None
    data_exploration_config: DataExplorationConfig | None = None


class InFieldLocationConfigRequest(WrappedInstanceListRequest, InFieldLocationConfig):
    def dump_instances(self) -> list[dict[str, Any]]:
        space: str | None = None
        external_id: str | None = None
        if self.data_exploration_config:
            space = self.data_exploration_config.space or self.space
            if self.data_exploration_config.external_id:
                external_id = self.data_exploration_config.external_id
            else:
                candidate = f"{self.external_id}_data_exploration_config"
                external_id = sanitize_instance_external_id(candidate)

        properties = self.model_dump(
            by_alias=True,
            exclude_unset=True,
            exclude={"data_exploration_config", "instance_type", "space", "external_id"},
        )
        if space and external_id:
            properties["dataExplorationConfig"] = {"space": space, "externalId": external_id}
        output: list[dict[str, Any]] = [
            {
                "instanceType": self.instance_type,
                "space": self.space,
                "externalId": self.external_id,
                "sources": [
                    {
                        "source": self.VIEW_ID.dump(),
                        "properties": properties,
                    }
                ],
            }
        ]
        if space and external_id and self.data_exploration_config:
            output.append(
                {
                    "instanceType": "node",
                    "space": space,
                    "externalId": external_id,
                    "sources": [
                        {
                            "source": DataExplorationConfig.VIEW_ID.dump(),
                            "properties": self.data_exploration_config.model_dump(
                                by_alias=True, exclude_unset=True, exclude={"space", "external_id"}
                            ),
                        }
                    ],
                }
            )
        return output

    def as_ids(self) -> list[TypedInstanceIdentifier]:
        output: list[TypedInstanceIdentifier] = [self.as_id()]
        if (
            self.data_exploration_config
            and self.data_exploration_config.space
            and self.data_exploration_config.external_id
        ):
            output.append(
                TypedNodeIdentifier(
                    space=self.data_exploration_config.space,
                    external_id=self.data_exploration_config.external_id,
                )
            )
        return output


class InFieldLocationConfigResponse(WrappedInstanceListResponse, InFieldLocationConfig):
    @classmethod
    def request_cls(cls) -> type[InFieldLocationConfigRequest]:
        return InFieldLocationConfigRequest

    def as_ids(self) -> list[TypedInstanceIdentifier]:
        output: list[TypedInstanceIdentifier] = [TypedNodeIdentifier(space=self.space, external_id=self.external_id)]
        if (
            self.data_exploration_config
            and self.data_exploration_config.space
            and self.data_exploration_config.external_id
        ):
            output.append(
                TypedNodeIdentifier(
                    space=self.data_exploration_config.space,
                    external_id=self.data_exploration_config.external_id,
                )
            )
        return output


class InFieldCDMLocationConfig(BaseModelObject):
    name: str | None = None
    description: str | None = None
    feature_toggles: dict[str, JsonValue] | None = None
    access_management: dict[str, JsonValue] | None = None
    data_filters: dict[str, JsonValue] | None = None
    data_storage: dict[str, JsonValue] | None = None
    view_mappings: dict[str, JsonValue] | None = None
    disciplines: list[dict[str, JsonValue]] | None = None
    data_exploration_config: dict[str, JsonValue] | None = None


class InFieldCDMLocationConfigRequest(WrappedInstanceRequest, InFieldCDMLocationConfig):
    VIEW_ID: ClassVar[TypedViewReference] = INFIELD_CDM_LOCATION_CONFIG_VIEW_ID
    instance_type: Literal["node"] = "node"

    def as_id(self) -> TypedNodeIdentifier:
        return TypedNodeIdentifier(space=self.space, external_id=self.external_id)


class InFieldCDMLocationConfigResponse(
    WrappedInstanceResponse[InFieldCDMLocationConfigRequest], InFieldCDMLocationConfig
):
    VIEW_ID: ClassVar[TypedViewReference] = INFIELD_CDM_LOCATION_CONFIG_VIEW_ID
    instance_type: Literal["node"] = "node"

    @classmethod
    def request_cls(cls) -> type[InFieldCDMLocationConfigRequest]:
        return InFieldCDMLocationConfigRequest

    def as_request_resource(self) -> InFieldCDMLocationConfigRequest:
        return InFieldCDMLocationConfigRequest.model_validate(self.dump(context="toolkit"), extra="ignore")



import sys
from pathlib import Path

from pydantic import BaseModel

from cognite_toolkit._cdf_tk.utils.file import read_yaml_content, read_yaml_file

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class YAMLFile(BaseModel):
    @classmethod
    def from_yaml(cls, yaml_str: str) -> Self:
        """Create an instance of the class from a YAML string."""
        data = read_yaml_content(yaml_str)
        return cls.model_validate(data)

    @classmethod
    def from_yaml_file(cls, file_path: Path) -> Self:
        """Create an instance of the class from a YAML file."""
        data = read_yaml_file(file_path, expected_output="dict")
        return cls.model_validate(data)



from collections.abc import Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.streams import StreamRequest, StreamResponse


class StreamsAPI(CDFResourceAPI[StreamResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/streams", item_limit=1),
                "delete": Endpoint(method="POST", path="/streams/delete", item_limit=1),
                "retrieve": Endpoint(method="GET", path="/streams/{streamId}", item_limit=1),
                "list": Endpoint(method="GET", path="/streams", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[StreamResponse]:
        return PagedResponse[StreamResponse].model_validate_json(response.body)

    def create(self, items: Sequence[StreamRequest]) -> list[StreamResponse]:
        """Create one or more streams.

        Args:
            items: Sequence of StreamRequest items to create.

        Returns:
            List of created StreamResponse items.
        """
        return self._request_item_response(items, "create")

    def delete(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete streams using their external IDs.

        Args:
            items: Sequence of ExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        if ignore_unknown_ids:
            # The endpoint does not support ignoreUnknownIds, so we have to do it on the client side
            return self._request_item_split_retries_no_response(items, "delete")
        else:
            return self._request_no_response(items, "delete")

    def retrieve(
        self, items: Sequence[ExternalId], include_statistics: bool = False, ignore_unknown_ids: bool = False
    ) -> list[StreamResponse]:
        """Retrieve streams by their external IDs.

        Note: The streams API only supports retrieving one stream at a time via path parameter.

        Args:
            items: Sequence of ExternalId objects to retrieve.
            include_statistics: Whether to include usage statistics in the response.
            ignore_unknown_ids: Whether to ignore unknown IDs.

        Returns:
            List of StreamResponse items.
        """
        results: list[StreamResponse] = []
        endpoint = self._method_endpoint_map["retrieve"]
        for item in items:
            response = self._http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=self._make_url(endpoint.path.format(streamId=item.external_id)),
                    method=endpoint.method,
                    parameters={"includeStatistics": include_statistics},
                )
            )
            if isinstance(response, SuccessResponse):
                results.append(StreamResponse.model_validate(response.body_json))
            elif ignore_unknown_ids:
                continue
            _ = response.get_success_or_raise()
        return results

    def list(self) -> list[StreamResponse]:
        """List all streams.

        Note: The streams API does not support pagination.

        Returns:
            List of StreamResponse items.
        """
        return self._list(limit=None)



"""Functions API for managing CDF functions.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Functions/operation/postFunctions
"""

from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.api.function_schedules import FunctionSchedulesAPI
from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.function import FunctionRequest, FunctionResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalOrExternalId


class FunctionsAPI(CDFResourceAPI[FunctionResponse]):
    """API for managing CDF functions.

    Note: Functions do not support update operations.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/functions", item_limit=1),
                "retrieve": Endpoint(method="POST", path="/functions/byids", item_limit=10),
                "delete": Endpoint(method="POST", path="/functions/delete", item_limit=10),
                "list": Endpoint(method="POST", path="/functions/list", item_limit=1000),
            },
        )
        self.schedules = FunctionSchedulesAPI(http_client)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[FunctionResponse]:
        return PagedResponse[FunctionResponse].model_validate_json(response.body)

    def create(self, items: Sequence[FunctionRequest]) -> list[FunctionResponse]:
        """Create functions in CDF.

        Args:
            items: List of FunctionRequest objects to create.

        Returns:
            List of created FunctionResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False
    ) -> list[FunctionResponse]:
        """Retrieve functions from CDF by ID.

        Args:
            items: List of ExternalId or InternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.

        Returns:
            List of retrieved FunctionResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def delete(self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete functions from CDF.

        Args:
            items: List of ExternalId or InternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def paginate(
        self,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[FunctionResponse]:
        """Get a page of functions from CDF.

        Args:
            limit: Maximum number of functions to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of FunctionResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
        )

    def iterate(
        self,
        limit: int | None = None,
    ) -> Iterable[list[FunctionResponse]]:
        """Iterate over all functions in CDF.

        Args:
            limit: Maximum total number of functions to return.

        Returns:
            Iterable of lists of FunctionResponse objects.
        """
        return self._iterate(limit=limit)

    def list(self, limit: int | None = None) -> list[FunctionResponse]:
        """List all functions in CDF.

        Args:
            limit: Maximum total number of functions to return.

        Returns:
            List of FunctionResponse objects.
        """
        return self._list(limit=limit)



import builtins
from typing import ClassVar, Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
)

from ._auth import BasicAuthenticationRequest, BasicAuthenticationResponse
from ._base import SourceRequestDefinition, SourceResponseDefinition
from ._certificate import AuthCertificateRequest, CACertificateRequest, CertificateResponse


class MQTTSource(BaseModelObject):
    type: Literal["mqtt5", "mqtt3"]
    host: str
    port: int | None = None
    use_tls: bool | None = None


class MQTTSourceRequest(MQTTSource, SourceRequestDefinition):
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"use_tls"})
    authentication: BasicAuthenticationRequest | None = None
    ca_certificate: CACertificateRequest | None = None
    auth_certificate: AuthCertificateRequest | None = None


class MQTTSourceResponse(
    SourceResponseDefinition,
    MQTTSource,
    ResponseResource[MQTTSourceRequest],
):
    authentication: BasicAuthenticationResponse | None = None
    ca_certificate: CertificateResponse | None = None
    auth_certificate: CertificateResponse | None = None

    @classmethod
    def request_cls(cls) -> builtins.type[MQTTSourceRequest]:
        return MQTTSourceRequest



import questionary

custom_style_fancy = questionary.Style(
    [
        ("qmark", "fg:#673ab7"),  # token in front of the question
        ("question", "bold"),  # question text
        ("answer", "fg:#f44336 bold"),  # submitted answer text behind the question
        ("pointer", "fg:#673ab7 bold"),  # pointer used in select and checkbox prompts
        ("highlighted", "fg:#673ab7 bold"),  # pointed-at choice in select and checkbox prompts
        ("selected", "fg:#673ab7"),  # style for a selected item of a checkbox
        ("separator", "fg:#cc5454"),  # separator in lists
        ("instruction", ""),  # user instructions for select, rawselect, checkbox
        ("text", ""),  # plain text
        ("disabled", "fg:#858585 italic"),  # disabled choices for select and checkbox prompts
    ]
)



from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.search_config import (
    SearchConfigRequest,
    SearchConfigResponse,
)


class SearchConfigurationsAPI(CDFResourceAPI[SearchConfigResponse]):
    """API for managing Search Configurations using the CDFResourceAPI pattern.

    This API manages search view configurations at:
    /apps/v1/projects/{project}/storage/config/apps/search/views
    """

    BASE_PATH = "/storage/config/apps/search/views"

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "upsert": Endpoint(method="POST", path=f"{self.BASE_PATH}/upsert", item_limit=1),
                "list": Endpoint(method="POST", path=f"{self.BASE_PATH}/list", item_limit=1000),
            },
        )

    def _make_url(self, path: str = "") -> str:
        """Create the full URL for this resource endpoint using the apps URL format."""
        return self._http_client.config.create_app_url(path)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[SearchConfigResponse]:
        return PagedResponse[SearchConfigResponse].model_validate_json(response.body)

    def create(self, items: Sequence[SearchConfigRequest]) -> list[SearchConfigResponse]:
        """Create or update a search configurations.

        Args:
            items: The search configuration to create or update.

        Returns:
            The created or updated search configurations.
        """
        endpoint = self._method_endpoint_map["upsert"]
        results: list[SearchConfigResponse] = []
        for item in items:
            request = RequestMessage(
                endpoint_url=self._make_url(endpoint.path),
                method=endpoint.method,
                body_content=item.model_dump(mode="json", by_alias=True),
            )
            result = self._http_client.request_single_retries(request)
            response = result.get_success_or_raise()
            results.append(SearchConfigResponse.model_validate_json(response.body))
        return results

    def update(self, items: Sequence[SearchConfigRequest]) -> list[SearchConfigResponse]:
        """Update a search configurations.

        Args:
            items: The search configuration to update.

        Returns:
            The updated search configurations.
        """
        return self.create(items)

    def paginate(self) -> PagedResponse[SearchConfigResponse]:
        """Get a single page of search configurations.

        Returns:
            PagedResponse of SearchConfigResponse objects.
        """
        return self._paginate(cursor=None, limit=100)

    def iterate(self) -> Iterable[list[SearchConfigResponse]]:
        """Iterate over all search configurations.

        Returns:
            Iterable of lists of SearchConfigResponse objects.
        """
        return self._iterate(limit=None)

    def list(self) -> list[SearchConfigResponse]:
        """List all search configurations.


        Returns:
            List of SearchConfigResponse objects.
        """
        return self._list(limit=None)



from collections import defaultdict
from collections.abc import Iterable
from itertools import groupby
from pathlib import Path
from typing import Any, cast

from cognite_toolkit._cdf_tk.commands.build_v2.data_classes import (
    AbsoluteDirPath,
    InsightList,
    ModelSyntaxError,
    ModuleSource,
    RelativeDirPath,
    RelativeFilePath,
)
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._module import BuildVariable
from cognite_toolkit._cdf_tk.constants import EXCL_FILES, MODULES
from cognite_toolkit._cdf_tk.cruds import CRUDS_BY_FOLDER_NAME_INCLUDE_ALPHA, ResourceTypes


class ModuleSourceParser:
    MODULE_ERROR_CODE = "MOD_001"
    VARIABLE_ERROR_CODE = "CONFIG_VARIABLE_001"

    def __init__(self, selected_modules: set[RelativeDirPath | str], organization_dir: AbsoluteDirPath) -> None:
        self.selected_modules = selected_modules
        self.organization_dir = organization_dir
        self.errors = InsightList()

    def parse(self, yaml_files: list[RelativeFilePath], variables: dict[str, Any]) -> list[ModuleSource]:
        source_by_module_id, orphans = self._find_modules(yaml_files, self.organization_dir)
        module_ids = list(source_by_module_id.keys())
        errors = self._validate_modules(module_ids, self.selected_modules, orphans)
        if errors:
            self.errors.extend(errors)
            return []
        selected_modules = self._select_modules(module_ids, self.selected_modules)
        build_variables, errors = self._parse_variables(variables, set(module_ids), set(selected_modules))
        if errors:
            self.errors.extend(errors)
            return []
        return self._create_module_soruces(selected_modules, source_by_module_id, build_variables)

    def _create_module_soruces(
        self,
        selected_modules: list[Path],
        source_by_module_id: dict[Path, ModuleSource],
        build_variables: dict[Path, list[BuildVariable]],
    ) -> list[ModuleSource]:
        module_sources: list[ModuleSource] = []
        for module in selected_modules:
            source = source_by_module_id[module]
            module_build_variables = build_variables.get(module, [])
            if module_build_variables:
                for iteration, module_variable in enumerate(module_build_variables, start=1):
                    module_sources.append(
                        source.model_copy(update={"variables": module_variable, "iteration": iteration})
                    )
            else:
                module_sources.append(source)
        return module_sources

    @classmethod
    def _find_modules(
        cls, yaml_files: list[RelativeFilePath], organization_dir: Path
    ) -> tuple[dict[RelativeDirPath, ModuleSource], list[RelativeDirPath]]:
        """Organizes YAML files by their module (top-level folder in the modules directory)."""
        source_by_module_id: dict[RelativeDirPath, ModuleSource] = {}
        orphan_files: list[RelativeDirPath] = []
        for yaml_file in yaml_files:
            if yaml_file.name in EXCL_FILES:
                continue
            relative_module_path, resource_folder = cls._get_module_path_from_resource_file_path(yaml_file)
            if relative_module_path and resource_folder:
                if relative_module_path not in source_by_module_id:
                    source_by_module_id[relative_module_path] = ModuleSource(
                        path=organization_dir / relative_module_path,
                        id=relative_module_path,
                    )
                source = source_by_module_id[relative_module_path]
                if resource_folder not in source.resource_files_by_folder:
                    source.resource_files_by_folder[resource_folder] = []
                source.resource_files_by_folder[resource_folder].append(organization_dir / yaml_file)
            else:
                orphan_files.append(yaml_file)
        return source_by_module_id, orphan_files

    @staticmethod
    def _get_module_path_from_resource_file_path(resource_file: Path) -> tuple[Path | None, ResourceTypes | None]:
        for parent in resource_file.parents:
            if parent.name in CRUDS_BY_FOLDER_NAME_INCLUDE_ALPHA:
                # We know that all keys in CRUDS_BY_FOLDER_NAME_INCLUDE_ALPHA are valid ResourceTypes,
                # so this cast is safe.
                return parent.parent, cast(ResourceTypes, parent.name)
        return None, None

    @classmethod
    def _validate_modules(
        cls,
        module_paths: list[RelativeDirPath],
        selection: set[RelativeDirPath | str],
        orphan_files: list[RelativeDirPath],
    ) -> list[ModelSyntaxError]:
        # Todo: Check the following
        # 1) If any module is inside another module. -> Error.
        # 2) If any selection by name is ambiguous (matches multiple found modules) -> Error.
        # 3) Any selected module by name that does not match any found module -> Error.
        # 4) If any orphan files that have a resource kind set are found -> Error.
        return []

    @classmethod
    def _select_modules(
        cls, module_paths: Iterable[RelativeDirPath], selection: set[RelativeDirPath | str]
    ) -> list[RelativeDirPath]:
        return [
            module_path
            for module_path in module_paths
            if module_path in selection
            or module_path.name in selection
            or any(parent in selection for parent in module_path.parents)
        ]

    @classmethod
    def _parse_module_variables(
        cls,
        variables: dict[str, Any],
        available_modules: set[RelativeDirPath],
        selected_modules: set[RelativeDirPath],
    ) -> tuple[dict[RelativeDirPath, dict[int, list[BuildVariable]]], list[ModelSyntaxError]]:
        all_available_paths = (
            {Path("")} | available_modules | {parent for module in available_modules for parent in module.parents}
        )
        selected_paths = (
            {Path("")} | selected_modules | {parent for module in selected_modules for parent in module.parents}
        )
        parsed_variables, errors = cls._parse_variables(variables, all_available_paths, selected_paths)
        variable_by_module = cls._organize_variables_by_module(parsed_variables, selected_modules)
        return variable_by_module, errors

    @classmethod
    def _parse_variables(
        cls, variables: dict[str, Any], available_paths: set[RelativeDirPath], selected_paths: set[RelativeDirPath]
    ) -> tuple[dict[RelativeDirPath, list[BuildVariable]], list[ModelSyntaxError]]:
        variables_by_path: dict[RelativeDirPath, list[BuildVariable]] = defaultdict(list)
        errors: list[ModelSyntaxError] = []
        to_check: list[tuple[RelativeDirPath, int | None, dict[str, Any]]] = [(Path(""), None, variables)]
        while to_check:
            path, iteration, subdict = to_check.pop()
            for key, value in subdict.items():
                subpath = path / key
                if isinstance(value, str | float | int | bool):
                    variables_by_path[path].append(
                        BuildVariable(id=subpath, value=value, is_selected=path in selected_paths, iteration=iteration)
                    )
                elif isinstance(value, dict):
                    if subpath in available_paths:
                        to_check.append((subpath, iteration, value))
                    else:
                        errors.append(
                            ModelSyntaxError(
                                code=cls.VARIABLE_ERROR_CODE,
                                message=f"Invalid variable path: {'.'.join(subpath.parts)}. This does not correspond to the "
                                f"folder structure inside the {MODULES} directory.",
                                fix="Ensure that the variable paths correspond to the folder structure inside the modules directory.",
                            )
                        )
                elif isinstance(value, list):
                    if all(isinstance(item, str | float | int | bool) for item in value):
                        variables_by_path[path].append(
                            BuildVariable(
                                id=subpath, value=value, is_selected=path in selected_paths, iteration=iteration
                            )
                        )
                    elif all(isinstance(item, dict) for item in value):
                        for idx, item in enumerate(value, start=1):
                            to_check.append((subpath, idx, item))
                    else:
                        errors.append(
                            ModelSyntaxError(
                                code=cls.VARIABLE_ERROR_CODE,
                                message=f"Invalid variable type in list for variable {'.'.join(subpath.parts)}.",
                                fix="Ensure that all items in the list are of the same supported type either (str, int, float, bool) or dict.",
                            )
                        )
                else:
                    raise NotImplementedError(f"Unsupported variable type: {type(value)} for variable {subpath}")
        return variables_by_path, errors

    @classmethod
    def _organize_variables_by_module(
        cls, variables_by_path: dict[RelativeDirPath, list[BuildVariable]], selected_modules: set[RelativeDirPath]
    ) -> dict[RelativeDirPath, dict[int, list[BuildVariable]]]:
        module_path_by_relative_paths: dict[frozenset[RelativeDirPath], RelativeDirPath] = {
            frozenset([module, *list(module.parents)]): module for module in selected_modules
        }
        variables_by_module: dict[RelativeDirPath, dict[int, list[BuildVariable]]] = defaultdict(
            lambda: defaultdict(list)
        )
        for variable_path, variables in variables_by_path.items():
            for module_paths, module in module_path_by_relative_paths.items():
                if variable_path in module_paths:
                    for iteration, variable in groupby(
                        sorted(variables, key=lambda v: v.iteration or 0), key=lambda v: v.iteration or 0
                    ):
                        variables_by_module[module][iteration or 0].extend(variable)
        return dict(variables_by_module)



import builtins
from datetime import datetime
from typing import ClassVar, Literal

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import (
    NodeReference,
    TypedNodeIdentifier,
    TypedViewReference,
    WrappedInstanceRequest,
    WrappedInstanceResponse,
)

COGNITE_FILE_VIEW_ID = TypedViewReference(space="cdf_cdm", external_id="CogniteFile", version="v1")


class CogniteFile(BaseModelObject):
    """Base class for CogniteFile containing common properties."""

    name: str | None = None
    description: str | None = None
    tags: list[str] | None = None
    aliases: list[str] | None = None
    source_id: str | None = None
    source_context: str | None = None
    source: NodeReference | None = None
    source_created_time: datetime | None = None
    source_updated_time: datetime | None = None
    source_created_user: str | None = None
    source_updated_user: str | None = None
    assets: list[NodeReference] | None = None
    mime_type: str | None = None
    directory: str | None = None
    category: NodeReference | None = None
    type: NodeReference | None = None


class CogniteFileRequest(WrappedInstanceRequest, CogniteFile):
    """CogniteFile request resource for creating/updating nodes."""

    VIEW_ID: ClassVar[TypedViewReference] = COGNITE_FILE_VIEW_ID
    instance_type: Literal["node"] = "node"
    space: str
    external_id: str

    def as_id(self) -> TypedNodeIdentifier:
        return TypedNodeIdentifier(space=self.space, external_id=self.external_id)


class CogniteFileResponse(WrappedInstanceResponse[CogniteFileRequest], CogniteFile):
    """CogniteFile response resource returned from API."""

    VIEW_ID: ClassVar[TypedViewReference] = COGNITE_FILE_VIEW_ID
    instance_type: Literal["node"] = "node"
    space: str
    external_id: str
    is_uploaded: bool | None = None
    uploaded_time: datetime | None = None

    @classmethod
    def request_cls(cls) -> builtins.type[CogniteFileRequest]:
        return CogniteFileRequest

    def as_request_resource(self) -> CogniteFileRequest:
        return CogniteFileRequest.model_validate(
            self.model_dump(mode="json", by_alias=True, exclude_unset=True, exclude={"instance_type"}),
            extra="ignore",
        )



from typing import Annotated, Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import DataProductVersionId, SemanticVersion
from cognite_toolkit._cdf_tk.constants import SPACE_FORMAT_PATTERN

from .base import BaseModelResource, ToolkitResource

SpaceId = Annotated[str, Field(pattern=SPACE_FORMAT_PATTERN, max_length=43)]


class ViewInstanceSpaces(BaseModelResource):
    read: list[SpaceId] = Field(
        default_factory=list, description="Instance space IDs from which you can read through this view."
    )
    write: list[SpaceId] = Field(
        default_factory=list, description="Instance space IDs to which you can write through this view."
    )


class DataProductVersionView(BaseModelResource):
    external_id: str = Field(description="External ID of the view in the data model.")
    instance_spaces: ViewInstanceSpaces = Field(
        default_factory=ViewInstanceSpaces, description="Instance spaces for this view."
    )


class DataProductVersionDataModel(BaseModelResource):
    external_id: str = Field(description="External ID of the referenced data model.")
    version: str = Field(description="Version of the referenced data model.")
    views: list[DataProductVersionView] = Field(
        default_factory=list,
        description="List of views with their instance spaces.",
        max_length=100,
    )


class DataProductVersionTerms(BaseModelResource):
    usage: str | None = Field(
        default=None, description="Permitted usage terms and conditions (markdown).", max_length=2000
    )
    limitations: str | None = Field(
        default=None, description="Usage limitations and restrictions (markdown).", max_length=500
    )


class DataProductVersionYAML(ToolkitResource):
    data_product_external_id: str = Field(
        description="External ID of the parent data product.",
        min_length=1,
        max_length=100,
        pattern=r"^[a-z]([a-z0-9_-]{0,98}[a-z0-9])?$",
    )
    version: SemanticVersion = Field(
        description="Semantic version of this data product version (major.minor.patch).",
    )
    data_model: DataProductVersionDataModel = Field(
        description="Immutable reference to the data model version associated with this data product version.",
    )
    status: Literal["draft", "published", "deprecated"] = Field(
        default="draft",
        description="The status of this data product version.",
    )
    description: str | None = Field(
        default=None,
        description="A detailed description of this version (markdown).",
        max_length=2000,
    )
    terms: DataProductVersionTerms | None = Field(
        default=None,
        description="Terms and conditions for using this data product version.",
    )

    def as_id(self) -> DataProductVersionId:
        return DataProductVersionId(
            data_product_external_id=self.data_product_external_id,
            version=self.version,
        )



from pydantic import Field, JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    Identifier,
    RequestResource,
    ResponseResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalId


class FunctionScheduleId(Identifier):
    """Identifier for a function schedule."""

    function_external_id: str
    name: str

    def __str__(self) -> str:
        return f"functionExternalId='{self.function_external_id}', name='{self.name}'"


class FunctionScheduleCredentials(BaseModelObject):
    """Credentials for function schedule authentication."""

    client_id: str
    client_secret: str


class FunctionSchedule(BaseModelObject):
    """Base class for function schedule with common fields."""

    name: str
    cron_expression: str
    description: str | None = None
    data: dict[str, JsonValue] | None = None


class FunctionScheduleRequest(FunctionSchedule, RequestResource):
    """Request resource for creating/updating function schedules."""

    # The 'id' field is not part of the request when creating a new resource,
    # but is needed when deleting an existing resource.
    id: int | None = Field(default=None, exclude=True)
    # function_id is set at creation time after looking up the function by external_id
    function_id: int | None = None
    function_external_id: str | None = Field(None, exclude=True)
    # nonce is set at creation time after creating a session
    nonce: str | None = None

    def as_id(self) -> InternalId:
        if self.id is None:
            raise ValueError("Cannot convert FunctionScheduleRequest to InternalId when id is None")
        return InternalId(id=self.id)


class FunctionScheduleResponse(FunctionSchedule, ResponseResource[FunctionScheduleRequest]):
    """Response resource for function schedules."""

    id: int
    created_time: int
    when: str
    function_id: int | None = None
    function_external_id: str | None = None
    session_id: int | None = None

    @classmethod
    def request_cls(cls) -> type[FunctionScheduleRequest]:
        return FunctionScheduleRequest
